import json

import pytest

from conan.api.model import RecipeReference
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_graph_build_order_override_error():
    """
    libc -> libb -> liba -> zlib/1.2
              |--------------/
      |-----override------> zlib/1.3
    """
    c = TestClient(light=True)
    c.save({"zlib/conanfile.py": GenConanfile("zlib"),
            "liba/conanfile.py": GenConanfile("liba", "0.1").with_requires("zlib/1.0"),
            "libb/conanfile.py": GenConanfile("libb", "0.1").with_requires("liba/0.1", "zlib/2.0"),
            "libc/conanfile.py": GenConanfile("libc", "0.1").with_requirement("libb/0.1")
                                                            .with_requirement("zlib/3.0",
                                                                              override=True)
            })
    c.run("export zlib --version=2.0")
    c.run("export zlib --version=3.0")
    c.run("export liba")
    c.run("export libb")
    c.run("export libc")
    c.run("graph info --requires=libc/0.1 --lockfile-out=output.lock")

    c.run("graph build-order --requires=libc/0.1 --lockfile=output.lock --order-by=configuration "
          "--build=missing --format=json")

    to_build = json.loads(c.stdout)
    for level in to_build["order"]:
        for package in level:
            binary = package["binary"]
            assert binary == "Build"  # All nodes in this case have to be built
            build_args = package["build_args"]
            c.run(f"install {build_args} --lockfile=output.lock")
            ref = RecipeReference.loads(package["ref"])
            assert f"{ref}: Building from source"

    c.run("install --requires=libc/0.1 --lockfile=output.lock")
    # All works, all binaries exist now
    assert "zlib/3.0: Already installed!" in c.out
    assert "liba/0.1: Already installed!" in c.out
    assert "libb/0.1: Already installed!" in c.out
    assert "libc/0.1: Already installed!" in c.out


@pytest.mark.parametrize("replace_pattern", ["*", "3.0"])
def test_graph_build_order_override_replace_requires(replace_pattern):
    """
    libc -> libb -> liba -> zlib/1.2
              |--------------/
      |-----override------> zlib/1.3

    replace_requires zlib -> zlib/system
    """
    c = TestClient(light=True)
    c.save({"zlib/conanfile.py": GenConanfile("zlib"),
            "liba/conanfile.py": GenConanfile("liba", "0.1").with_requires("zlib/1.0"),
            "libb/conanfile.py": GenConanfile("libb", "0.1").with_requires("liba/0.1", "zlib/2.0"),
            "libc/conanfile.py": GenConanfile("libc", "0.1").with_requirement("libb/0.1")
                                                            .with_requirement("zlib/3.0",
                                                                              override=True),
            "profile": f"[replace_requires]\nzlib/{replace_pattern}: zlib/system"
            })
    c.run("export zlib --version=2.0")
    c.run("export zlib --version=3.0")
    c.run("export zlib --version=system")
    c.run("export liba")
    c.run("export libb")
    c.run("export libc")
    c.run("lock create --requires=libc/0.1 --lockfile-out=output.lock -pr=profile")

    c.run("graph build-order --requires=libc/0.1 --lockfile=output.lock --order-by=configuration "
          "--build=missing -pr=profile --format=json")

    to_build = json.loads(c.stdout)
    for level in to_build["order"]:
        for package in level:
            binary = package["binary"]
            assert binary == "Build"  # All nodes in this case have to be built
            build_args = package["build_args"]
            c.run(f"install {build_args} --lockfile=output.lock -pr=profile")
            ref = RecipeReference.loads(package["ref"])
            assert f"{ref}: Building from source"

    c.run("install --requires=libc/0.1 --lockfile=output.lock -pr=profile")
    # All works, all binaries exist now
    assert "zlib/system: Already installed!" in c.out
    assert "liba/0.1: Already installed!" in c.out
    assert "libb/0.1: Already installed!" in c.out
    assert "libc/0.1: Already installed!" in c.out


def test_single_config_decentralized_overrides():
    r""" same scenario as "test_single_config_centralized()", but distributing the build in
    different build servers, using the "build-order"
    Now with overrides

    pkga -> toola/1.0 -> toolb/1.0 -> toolc/1.0
                \------override-----> toolc/2.0
    pkgb -> toola/2.0 -> toolb/1.0 -> toolc/1.0
                \------override-----> toolc/3.0
    pkgc -> toola/3.0 -> toolb/1.0 -> toolc/1.0
    """
    c = TestClient(light=True)
    c.save({"toolc/conanfile.py": GenConanfile("toolc"),
            "toolb/conanfile.py": GenConanfile("toolb").with_requires("toolc/1.0"),
            "toola/conanfile.py": GenConanfile("toola", "1.0").with_requirement("toolb/1.0")
                                                              .with_requirement("toolc/2.0",
                                                                                override=True),
            "toola2/conanfile.py": GenConanfile("toola", "2.0").with_requirement("toolb/1.0")
                                                               .with_requirement("toolc/3.0",
                                                                                 override=True),
            "toola3/conanfile.py": GenConanfile("toola", "3.0").with_requirement("toolb/1.0"),
            "pkga/conanfile.py": GenConanfile("pkga", "1.0").with_tool_requires("toola/1.0"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "1.0").with_requires("pkga/1.0")
                                                            .with_tool_requires("toola/2.0"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "1.0").with_requires("pkgb/1.0")
                                                            .with_tool_requires("toola/3.0"),
            })
    c.run("export toolc --version=1.0")
    c.run("export toolc --version=2.0")
    c.run("export toolc --version=3.0")

    c.run("export toolb --version=1.0")

    c.run("export toola")
    c.run("export toola2")
    c.run("export toola3")

    c.run("export pkga")
    c.run("export pkgb")
    c.run("lock create pkgc")
    lock = json.loads(c.load("pkgc/conan.lock"))
    requires = "\n".join(lock["build_requires"])
    assert "toolc/3.0" in requires
    assert "toolc/2.0" in requires
    assert "toolc/1.0" in requires
    assert len(lock["overrides"]) == 1
    assert set(lock["overrides"]["toolc/1.0"]) == {"toolc/3.0", "toolc/2.0", None}

    c.run("graph build-order pkgc --lockfile=pkgc/conan.lock --format=json --build=missing")
    to_build = json.loads(c.stdout)
    for level in to_build:
        for elem in level:
            for package in elem["packages"][0]:  # assumes no dependencies between packages
                binary = package["binary"]
                assert binary == "Build"  # All nodes in this case have to be built
                build_args = package["build_args"]
                c.run(f"install {build_args} --lockfile=pkgc/conan.lock")

    c.run("install pkgc --lockfile=pkgc/conan.lock")
    # All works, all binaries exist now
    assert "pkga/1.0: Already installed!" in c.out
    assert "pkgb/1.0: Already installed!" in c.out


def test_single_config_decentralized_overrides_nested():
    r""" same scenario as "test_single_config_centralized()", but distributing the build in
    different build servers, using the "build-order"
    Now with overrides

    pkga -> toola/1.0 -> libb/1.0 -> libc/1.0 -> libd/1.0 -> libe/1.0 -> libf/1.0
               \                          \-----------override---------> libf/2.0
                \--------------------override--------------------------> libf/3.0
    """
    c = TestClient(light=True)
    c.save({"libf/conanfile.py": GenConanfile("libf"),
            "libe/conanfile.py": GenConanfile("libe", "1.0").with_requires("libf/1.0"),
            "libd/conanfile.py": GenConanfile("libd", "1.0").with_requires("libe/1.0"),
            "libc/conanfile.py": GenConanfile("libc", "1.0").with_requirement("libd/1.0")
                                                            .with_requirement("libf/2.0",
                                                                              override=True),
            "libb/conanfile.py": GenConanfile("libb", "1.0").with_requires("libc/1.0"),
            "toola/conanfile.py": GenConanfile("toola", "1.0").with_requirement("libb/1.0")
                                                              .with_requirement("libf/3.0",
                                                                                override=True),
            "pkga/conanfile.py": GenConanfile("pkga", "1.0").with_tool_requires("toola/1.0"),
            })

    c.run("export libf --version=3.0")
    c.run("export libe")
    c.run("export libd")
    c.run("export libc")
    c.run("export libb")
    c.run("export toola")

    c.run("lock create pkga")
    lock = json.loads(c.load("pkga/conan.lock"))
    assert lock["overrides"] == {"libf/1.0": ["libf/3.0"],
                                 "libf/2.0": ["libf/3.0"]}

    c.run("graph build-order pkga --lockfile=pkga/conan.lock --format=json --build=missing")
    to_build = json.loads(c.stdout)
    for level in to_build:
        for elem in level:
            ref = elem["ref"]
            if "libc" in ref:
                pass
            for package in elem["packages"][0]:  # assumes no dependencies between packages
                binary = package["binary"]
                assert binary == "Build"  # All nodes in this case have to be built
                build_args = package["build_args"]
                c.run(f"install {build_args} --lockfile=pkga/conan.lock")

    c.run("install pkga --lockfile=pkga/conan.lock")
    # All works, all binaries exist now
    assert "Install finished successfully" in c.out


@pytest.mark.parametrize("forced", [False, True])
def test_single_config_decentralized_overrides_multi(forced):
    r""" same scenario as "test_single_config_centralized()", but distributing the build in
    different build servers, using the "build-order"
    Now with overrides

    pkga -> toola/1.0 -> libb/1.0 -> libc/1.0 -> libd/1.0 -> libe/1.0 -> libf/1.0
      |         \                          \-----------override--------> libf/2.0
      |          \--------------------override-------------------------> libf/3.0
    pkgb -> toola/1.1 -> libb/1.0 -> libc/1.0 -> libd/1.0 -> libe/1.0 -> libf/1.0
      |         \                          \-----------override--------> libf/2.0
      |          \--------------------override-------------------------> libf/4.0
    pkgc -> toola/1.2 -> libb/1.0 -> libc/1.0 -> libd/1.0 -> libe/1.0 -> libf/1.0
                                           \-----------override--------> libf/2.0
    """
    override, force = (True, False) if not forced else (False, True)
    c = TestClient(light=True)
    c.save({"libf/conanfile.py": GenConanfile("libf"),
            "libe/conanfile.py": GenConanfile("libe", "1.0").with_requires("libf/1.0"),
            "libd/conanfile.py": GenConanfile("libd", "1.0").with_requires("libe/1.0"),
            "libc/conanfile.py": GenConanfile("libc", "1.0").with_requirement("libd/1.0")
                                                            .with_requirement("libf/2.0",
                                                                              override=override,
                                                                              force=force),
            "libb/conanfile.py": GenConanfile("libb", "1.0").with_requires("libc/1.0"),
            "toola/conanfile.py": GenConanfile("toola", "1.0").with_requirement("libb/1.0")
                                                              .with_requirement("libf/3.0",
                                                                                override=override,
                                                                                force=force),
            "toola1/conanfile.py": GenConanfile("toola", "1.1").with_requirement("libb/1.0")
                                                               .with_requirement("libf/4.0",
                                                                                 override=override,
                                                                                 force=force),
            "toola2/conanfile.py": GenConanfile("toola", "1.2").with_requirement("libb/1.0"),
            "pkga/conanfile.py": GenConanfile("pkga", "1.0").with_tool_requires("toola/1.0"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "1.0").with_requires("pkga/1.0")
                                                            .with_tool_requires("toola/1.1"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "1.0").with_requires("pkgb/1.0")
                                                            .with_tool_requires("toola/1.2"),
            })

    c.run("export libf --version=2.0")
    c.run("export libf --version=3.0")
    c.run("export libf --version=4.0")
    c.run("export libe")
    c.run("export libd")
    c.run("export libc")
    c.run("export libb")

    c.run("export toola")
    c.run("export toola1")
    c.run("export toola2")

    c.run("export pkga")
    c.run("export pkgb")
    c.run("lock create pkgc")
    lock = json.loads(c.load("pkgc/conan.lock"))
    assert len(lock["overrides"]) == 2
    assert set(lock["overrides"]["libf/1.0"]) == {"libf/4.0", "libf/2.0", "libf/3.0"}

    if forced:  # When forced, there is one libf/2.0 that is not overriden
        assert set(lock["overrides"]["libf/2.0"]) == {"libf/4.0", "libf/3.0", None}
    else:
        assert set(lock["overrides"]["libf/2.0"]) == {"libf/4.0", "libf/3.0"}

    c.run("graph build-order pkgc --lockfile=pkgc/conan.lock --format=json --build=missing")
    to_build = json.loads(c.stdout)
    for level in to_build:
        for elem in level:
            ref = elem["ref"]
            if "libc" in ref:
                pass
            for package in elem["packages"][0]:  # assumes no dependencies between packages
                binary = package["binary"]
                assert binary == "Build"  # All nodes in this case have to be built
                build_args = package["build_args"]
                c.run(f"install {build_args} --lockfile=pkgc/conan.lock")

    c.run("install pkgc --lockfile=pkgc/conan.lock")
    # All works, all binaries exist now
    assert "pkga/1.0: Already installed!" in c.out
    assert "pkgb/1.0: Already installed!" in c.out


@pytest.mark.parametrize("replace_pattern", ["*", "1.0", "2.0", "3.0", "4.0"])
@pytest.mark.parametrize("forced", [False, True])
def test_single_config_decentralized_overrides_multi_replace_requires(replace_pattern, forced):
    r""" same scenario as "test_single_config_centralized()", but distributing the build in
    different build servers, using the "build-order"
    Now with overrides

    pkga -> toola/1.0 -> libb/1.0 -> libc/1.0 -> libd/1.0 -> libe/1.0 -> libf/1.0
      |         \                          \-----------override--------> libf/2.0
      |          \--------------------override-------------------------> libf/3.0
    pkgb -> toola/1.1 -> libb/1.0 -> libc/1.0 -> libd/1.0 -> libe/1.0 -> libf/1.0
      |         \                          \-----------override--------> libf/2.0
      |          \--------------------override-------------------------> libf/4.0
    pkgc -> toola/1.2 -> libb/1.0 -> libc/1.0 -> libd/1.0 -> libe/1.0 -> libf/1.0
                                           \-----------override--------> libf/2.0
    """
    override, force = (True, False) if not forced else (False, True)
    c = TestClient(light=True)
    c.save({"libf/conanfile.py": GenConanfile("libf"),
            "libe/conanfile.py": GenConanfile("libe", "1.0").with_requires("libf/1.0"),
            "libd/conanfile.py": GenConanfile("libd", "1.0").with_requires("libe/1.0"),
            "libc/conanfile.py": GenConanfile("libc", "1.0").with_requirement("libd/1.0")
                                                            .with_requirement("libf/2.0",
                                                                              override=override,
                                                                              force=force),
            "libb/conanfile.py": GenConanfile("libb", "1.0").with_requires("libc/1.0"),
            "toola/conanfile.py": GenConanfile("toola", "1.0").with_requirement("libb/1.0")
                                                              .with_requirement("libf/3.0",
                                                                                override=override,
                                                                                force=force),
            "toola1/conanfile.py": GenConanfile("toola", "1.1").with_requirement("libb/1.0")
                                                               .with_requirement("libf/4.0",
                                                                                 override=override,
                                                                                 force=force),
            "toola2/conanfile.py": GenConanfile("toola", "1.2").with_requirement("libb/1.0"),
            "pkga/conanfile.py": GenConanfile("pkga", "1.0").with_tool_requires("toola/1.0"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "1.0").with_requires("pkga/1.0")
                                                            .with_tool_requires("toola/1.1"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "1.0").with_requires("pkgb/1.0")
                                                            .with_tool_requires("toola/1.2"),
            "profile": f"include(default)\n[replace_requires]\nlibf/{replace_pattern}: libf/system"
            })

    c.run("export libf --version=2.0")
    c.run("export libf --version=3.0")
    c.run("export libf --version=4.0")
    c.run("export libf --version=system")
    c.run("export libe")
    c.run("export libd")
    c.run("export libc")
    c.run("export libb")

    c.run("export toola")
    c.run("export toola1")
    c.run("export toola2")

    c.run("export pkga")
    c.run("export pkgb")
    c.run("lock create pkgc -pr:b=profile")

    # overrides will be different everytime, just checking that things can be built
    c.run("graph build-order pkgc --lockfile=pkgc/conan.lock --format=json -pr:b=profile "
          "--build=missing")

    to_build = json.loads(c.stdout)
    for level in to_build:
        for elem in level:
            ref = elem["ref"]
            if "libc" in ref:
                pass
            for package in elem["packages"][0]:  # assumes no dependencies between packages
                binary = package["binary"]
                assert binary == "Build"  # All nodes in this case have to be built
                build_args = package["build_args"]
                c.run(f"install {build_args} --lockfile=pkgc/conan.lock -pr:b=profile")

    c.run("install pkgc --lockfile=pkgc/conan.lock -pr:b=profile")
    # All works, all binaries exist now
    assert "pkga/1.0: Already installed!" in c.out
    assert "pkgb/1.0: Already installed!" in c.out
    if replace_pattern == "1.0":  # These will overriden by downstream
        assert "libf/system#7fb6d926dabeb955bcea1cafedf953c8 - Cache" not in c.out
    else:
        assert "libf/system#7fb6d926dabeb955bcea1cafedf953c8 - Cache" in c.out



import os
import textwrap

import pytest

from conan.internal.api.profile.profile_loader import _ProfileParser, ProfileLoader
from conan.errors import ConanException
from conan.api.model import RecipeReference
from conan.test.utils.mocks import ConanFileMock, MockSettings
from conan.test.utils.test_files import temp_folder
from conan.internal.util.files import save


def test_profile_parser():
    txt = textwrap.dedent(r"""
        include(a/path/to\profile.txt)
        include(other/path/to/file.txt)

        [settings]
        os=2
        """)
    a = _ProfileParser(txt)
    assert a.includes == [r"a/path/to\profile.txt", "other/path/to/file.txt"]
    assert a.profile_text == textwrap.dedent("""[settings]
os=2""")

    txt = ""
    a = _ProfileParser(txt)
    assert a.includes == []
    assert a.profile_text == ""

    txt = textwrap.dedent(r"""
        includes(a/path/to\profile.txt)
        """)

    with pytest.raises(ConanException):
        try:
            _ProfileParser(txt)
        except Exception as error:
            assert "Error while parsing line 1" in error.args[0]
            raise


def test_profiles_includes():
    tmp = temp_folder()

    def save_profile(txt, name):
        abs_profile_path = os.path.join(tmp, name)
        save(abs_profile_path, txt)

    os.mkdir(os.path.join(tmp, "subdir"))

    profile0 = textwrap.dedent("""
            {% set ROOTVAR=0 %}


            [tool_requires]
            one/1.{{ROOTVAR}}@lasote/stable
            two/1.2@lasote/stable
        """)
    save_profile(profile0, "subdir/profile0.txt")

    profile1 = textwrap.dedent("""
            # Include in subdir, curdir
            include(./profile0.txt)

            [settings]
            os=Windows
            [options]
            zlib*:aoption=1
            zlib*:otheroption=1
        """)

    save_profile(profile1, "subdir/profile1.txt")

    profile2 = textwrap.dedent("""
            #  Include in subdir
            {% set MYVAR=1 %}
            include(./subdir/profile1.txt)
            [settings]
            os={{MYVAR}}
        """)

    save_profile(profile2, "profile2.txt")
    profile3 = textwrap.dedent("""
            [tool_requires]
            one/1.5@lasote/stable
        """)

    save_profile(profile3, "profile3.txt")

    profile4 = textwrap.dedent("""
            include(./profile2.txt)
            include(./profile3.txt)

            [options]
            zlib*:otheroption=12
        """)

    save_profile(profile4, "profile4.txt")

    profile_loader = ProfileLoader(cache_folder=temp_folder())  # If not used cache, will not error
    profile = profile_loader.load_profile("./profile4.txt", tmp)

    assert profile.settings == {"os": "1"}
    assert profile.options["zlib*"].aoption == 1
    assert profile.options["zlib*"].otheroption == 12
    assert profile.tool_requires == {"*": [RecipeReference.loads("one/1.5@lasote/stable"),
                                           RecipeReference.loads("two/1.2@lasote/stable")]}


def test_profile_compose_platform_tool_requires():
    tmp = temp_folder()
    save(os.path.join(tmp, "profile0"), "[platform_tool_requires]\ntool1/1.0")
    save(os.path.join(tmp, "profile1"), "[platform_tool_requires]\ntool2/2.0")
    save(os.path.join(tmp, "profile2"), "include(./profile0)\n[platform_tool_requires]\ntool3/3.0")
    save(os.path.join(tmp, "profile3"), "include(./profile0)\n[platform_tool_requires]\ntool1/1.1")

    profile_loader = ProfileLoader(cache_folder=temp_folder())  # If not used cache, will not error
    profile2 = profile_loader.load_profile("./profile2", tmp)
    assert profile2.platform_tool_requires == [RecipeReference.loads("tool1/1.0"),
                                               RecipeReference.loads("tool3/3.0")]
    profile3 = profile_loader.load_profile("./profile3", tmp)
    assert profile3.platform_tool_requires == [RecipeReference.loads("tool1/1.1")]
    profile0 = profile_loader.load_profile("./profile0", tmp)
    profile1 = profile_loader.load_profile("./profile1", tmp)
    profile0.compose_profile(profile1)
    assert profile0.platform_tool_requires == [RecipeReference.loads("tool1/1.0"),
                                               RecipeReference.loads("tool2/2.0")]


def test_profile_compose_tool_requires():
    tmp = temp_folder()
    save(os.path.join(tmp, "profile0"), "[tool_requires]\ntool1/1.0")
    save(os.path.join(tmp, "profile1"), "[tool_requires]\ntool2/2.0")
    save(os.path.join(tmp, "profile2"), "include(./profile0)\n[tool_requires]\ntool3/3.0")
    save(os.path.join(tmp, "profile3"), "include(./profile0)\n[tool_requires]\ntool1/1.1")

    profile_loader = ProfileLoader(cache_folder=temp_folder())  # If not used cache, will not error
    profile2 = profile_loader.load_profile("./profile2", tmp)
    assert profile2.tool_requires == {"*": [RecipeReference.loads("tool1/1.0"),
                                            RecipeReference.loads("tool3/3.0")]}
    profile3 = profile_loader.load_profile("./profile3", tmp)
    assert profile3.tool_requires == {"*": [RecipeReference.loads("tool1/1.1")]}

    profile0 = profile_loader.load_profile("./profile0", tmp)
    profile1 = profile_loader.load_profile("./profile1", tmp)
    profile0.compose_profile(profile1)
    assert profile0.tool_requires == {"*": [RecipeReference.loads("tool1/1.0"),
                                            RecipeReference.loads("tool2/2.0")]}


def test_profile_include_order():
    tmp = temp_folder()

    save(os.path.join(tmp, "profile1.txt"), '{% set MYVAR="fromProfile1" %}')

    profile2 = textwrap.dedent("""
            include(./profile1.txt)
            {% set MYVAR="fromProfile2" %}
            [settings]
            os={{MYVAR}}
        """)
    save(os.path.join(tmp, "profile2.txt"), profile2)
    profile_loader = ProfileLoader(cache_folder=temp_folder())  # If not used cache, will not error
    profile = profile_loader.load_profile("./profile2.txt", tmp)

    assert profile.settings["os"] == "fromProfile2"


def test_profile_load_absolute_path():
    """ When passing absolute path as profile file, it MUST be used.
        read_profile(/abs/path/profile, /abs, /.conan/profiles)
        /abs/path/profile MUST be consumed as target profile
    """
    current_profile_folder = temp_folder()
    current_profile_path = os.path.join(current_profile_folder, "default")
    save(current_profile_path, "[settings]\nos=Windows")

    profile_loader = ProfileLoader(cache_folder=temp_folder())  # If not used cache, will not error
    profile = profile_loader.load_profile(current_profile_path)
    assert "Windows" == profile.settings["os"]


def test_profile_load_relative_path_dot():
    """ When passing relative ./path as profile file, it MUST be used
        read_profile(./profiles/profile, /tmp, /.conan/profiles)
        /tmp/profiles/profile MUST be consumed as target profile
    """
    current_profile_folder = temp_folder()
    current_profile_path = os.path.join(current_profile_folder, "profiles", "default")
    save(current_profile_path, "[settings]\nos=Windows")

    profile_loader = ProfileLoader(cache_folder=temp_folder())  # If not used cache, will not error
    profile = profile_loader.load_profile("./profiles/default", current_profile_folder)
    assert "Windows" == profile.settings["os"]


def test_profile_load_relative_path_pardir():
    """ When passing relative ../path as profile file, it MUST be used
        read_profile(../profiles/profile, /tmp/current, /.conan/profiles)
        /tmp/profiles/profile MUST be consumed as target profile
    """
    tmp = temp_folder()
    current_profile_path = os.path.join(tmp, "profiles", "default")
    cwd = os.path.join(tmp, "user_folder")
    save(current_profile_path, "[settings]\nos=Windows")

    profile_loader = ProfileLoader(cache_folder=temp_folder())  # If not used cache, will not error
    profile = profile_loader.load_profile("../profiles/default", cwd)
    assert "Windows" == profile.settings["os"]


def test_profile_buildenv():
    tmp = temp_folder()
    txt = textwrap.dedent("""
        [buildenv]
        MyVar1=My Value; 11
        MyVar1+=MyValue12
        MyPath1=(path)/some/path11
        MyPath1+=(path)/other path/path12
        """)
    current_profile_path = os.path.join(tmp, "default")
    save(current_profile_path, txt)

    profile_loader = ProfileLoader(cache_folder=temp_folder())  # If not used cache, will not error
    profile = profile_loader.load_profile(current_profile_path)
    buildenv = profile.buildenv
    env = buildenv.get_profile_env(None)
    conanfile = ConanFileMock()
    conanfile.settings_build = MockSettings({"os": "Linux", "arch": "x86_64"})
    env_vars = env.vars(conanfile)
    assert env_vars.get("MyVar1") == "My Value; 11 MyValue12"
    assert env_vars.get("MyPath1") == "/some/path11:/other path/path12"

    conanfile.settings_build = MockSettings({"os": "Windows", "arch": "x86_64"})
    env_vars = env.vars(conanfile)
    assert env_vars.get("MyPath1") == "/some/path11;/other path/path12"


@pytest.mark.parametrize("conf_name", [
    "core.gzip:compresslevel=5",
    "core.gzip:compresslevel"
])
def test_profile_core_confs_error(conf_name):
    tmp = temp_folder()
    current_profile_path = os.path.join(tmp, "default")
    save(current_profile_path, "")

    profile_loader = ProfileLoader(cache_folder=temp_folder())  # If not used cache, will not error

    with pytest.raises(ConanException) as exc:
        profile_loader.from_cli_args([], [], [], [conf_name], None)
    assert "[conf] 'core.*' configurations are not allowed in profiles" in str(exc.value)


def test_profile_compose_numbers():
    tmp = temp_folder()
    txt = textwrap.dedent("""
            [conf]
            user.version:value=8.1
            pkg/*:user.version:value=10
            """)
    current_profile_path = os.path.join(tmp, "default")
    save(current_profile_path, txt)

    profile_loader = ProfileLoader(cache_folder=temp_folder())  # If not used cache, will not error
    profile = profile_loader.load_profile(current_profile_path)
    assert profile.conf.get("user.version:value") == 8.1
    assert profile.conf.get("pkg/*:user.version:value") == 10



import os
import textwrap

import pytest

from conan.tools.files import load
from conan.tools.google import BazelToolchain
from conan.test.utils.tools import TestClient


@pytest.fixture(scope="module")
def conanfile():
    return textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.google import BazelToolchain, bazel_layout
        class ExampleConanIntegration(ConanFile):
            settings = "os", "arch", "build_type", "compiler"
            options = {"shared": [True, False], "fPIC": [True, False]}
            default_options = {"shared": False, "fPIC": True}
            generators = 'BazelToolchain'
            def layout(self):
                bazel_layout(self)
    """)


def test_default_bazel_toolchain(conanfile):
    profile = textwrap.dedent("""
    [settings]
    arch=x86_64
    build_type=Release
    compiler=apple-clang
    compiler.cppstd=gnu17
    compiler.libcxx=libc++
    compiler.version=13.0
    os=Macos
    """)

    c = TestClient()
    c.save({"conanfile.py": conanfile,
            "profile": profile})
    c.run("install . -pr profile")
    content = load(c, os.path.join(c.current_folder, "conan", BazelToolchain.bazelrc_name))
    assert "build:conan-config --cxxopt=-std=gnu++17" in content
    assert "build:conan-config --force_pic=True" in content
    assert "build:conan-config --dynamic_mode=off" in content
    assert "build:conan-config --compilation_mode=opt" in content


def test_bazel_toolchain_and_flags(conanfile):
    profile = textwrap.dedent("""
    [settings]
    arch=x86_64
    build_type=Release
    compiler=apple-clang
    compiler.cppstd=gnu17
    compiler.libcxx=libc++
    compiler.version=13.0
    os=Macos
    [options]
    shared=True
    [conf]
    tools.build:cxxflags=["--flag1", "--flag2"]
    tools.build:cflags+=["--flag3", "--flag4"]
    tools.build:sharedlinkflags+=["--flag5"]
    tools.build:exelinkflags+=["--flag6"]
    tools.build:linker_scripts+=["myscript.sh"]
    """)
    c = TestClient()
    c.save({"conanfile.py": conanfile,
            "profile": profile})
    c.run("install . -pr profile")
    content = load(c, os.path.join(c.current_folder, "conan", BazelToolchain.bazelrc_name))
    assert "build:conan-config --conlyopt=--flag3 --conlyopt=--flag4" in content
    assert "build:conan-config --cxxopt=-std=gnu++17 --cxxopt=--flag1 --cxxopt=--flag2" in content
    assert "build:conan-config --linkopt=--flag5 --linkopt=--flag6 --linkopt=-T'myscript.sh'" in content
    assert "build:conan-config --force_pic=True" not in content
    assert "build:conan-config --dynamic_mode=fully" in content
    assert "build:conan-config --compilation_mode=opt" in content


def test_bazel_toolchain_and_cross_compilation(conanfile):
    profile = textwrap.dedent("""
    [settings]
    arch=x86_64
    build_type=Release
    compiler=apple-clang
    compiler.cppstd=gnu17
    compiler.libcxx=libc++
    compiler.version=13.0
    os=Macos
    """)
    profile_host = textwrap.dedent("""
    [settings]
    arch=armv8
    build_type=Release
    compiler=apple-clang
    compiler.cppstd=gnu17
    compiler.libcxx=libc++
    compiler.version=13.0
    os=Macos

    """)
    c = TestClient()
    c.save({"conanfile.py": conanfile,
            "profile": profile,
            "profile_host": profile_host})
    c.run("install . -pr:b profile -pr:h profile_host")
    content = load(c, os.path.join(c.current_folder, "conan", BazelToolchain.bazelrc_name))
    assert "build:conan-config --cpu=darwin_arm64" in content


def test_toolchain_attributes_and_conf_priority():
    """
    Tests that all the attributes are appearing correctly in the conan_bzl.rc even defining
    some conf variables
    """
    profile = textwrap.dedent("""
    [settings]
    arch=x86_64
    build_type=Release
    compiler=apple-clang
    compiler.cppstd=gnu17
    compiler.libcxx=libc++
    compiler.version=13.0
    os=Macos
    [conf]
    tools.build:cxxflags=["--flag1"]
    tools.build:cflags+=["--flag3"]
    tools.build:sharedlinkflags+=["--linkflag5"]
    tools.build:exelinkflags+=["--linkflag6"]
    """)
    conanfile = textwrap.dedent("""
    from conan import ConanFile
    from conan.tools.google import BazelToolchain
    class ExampleConanIntegration(ConanFile):
        settings = "os", "arch", "build_type", "compiler"
        options = {"shared": [True, False], "fPIC": [True, False]}
        default_options = {"shared": False, "fPIC": True}

        def generate(self):
            bz = BazelToolchain(self)
            bz.copt = ["copt1"]
            bz.conlyopt = ["conly1"]
            bz.cxxopt = ["cxxopt1"]
            bz.linkopt = ["linkopt1"]
            bz.force_pic = True
            bz.dynamic_mode = "auto"
            bz.compilation_mode = "fastbuild"
            bz.compiler = "gcc"
            bz.cpu = "armv8"
            bz.crosstool_top = "my_crosstool"
            bz.generate()
    """)
    c = TestClient()
    c.save({"conanfile.py": conanfile,
            "profile": profile})
    c.run("install . -pr profile")
    content = load(c, os.path.join(c.current_folder, BazelToolchain.bazelrc_name))
    expected = textwrap.dedent("""\
    # Automatic bazelrc file created by Conan
    build:conan-config --copt=copt1
    build:conan-config --conlyopt=conly1 --conlyopt=--flag3
    build:conan-config --cxxopt=-std=gnu++17 --cxxopt=cxxopt1 --cxxopt=--flag1
    build:conan-config --linkopt=linkopt1 --linkopt=--linkflag5 --linkopt=--linkflag6
    build:conan-config --force_pic=True
    build:conan-config --dynamic_mode=auto
    build:conan-config --compilation_mode=fastbuild
    build:conan-config --compiler=gcc
    build:conan-config --cpu=armv8
    build:conan-config --crosstool_top=my_crosstool""")
    assert expected == content



import os
import re
import textwrap

import pytest

from conan.internal.cache.home_paths import HomePaths
from conan.test.utils.tools import TestClient, GenConanfile


class TestCacheCompatibles:
    @pytest.fixture()
    def client(self):
        client = TestClient()
        debug_compat = textwrap.dedent("""\
            def debug_compat(conanfile):
                result = []
                if conanfile.settings.build_type == "Debug":
                    result.append({"settings": [("build_type", "Release")]})
                return result
            """)
        compatibles = textwrap.dedent("""\
            from debug_compat import debug_compat
            def compatibility(conanfile):
                if conanfile.name != "dep":
                    return
                return debug_compat(conanfile)
            """)
        client.save_home({"extensions/plugins/compatibility/compatibility.py": compatibles,
                          "extensions/plugins/compatibility/debug_compat.py": debug_compat})
        return client

    def test_compatible_build_type(self, client):
        client.save({"dep/conanfile.py": GenConanfile("dep", "0.1").with_setting("build_type"),
                     "consumer/conanfile.py": GenConanfile().with_requires("dep/0.1")})

        client.run("create dep -s build_type=Release")
        package_id = client.created_package_id("dep/0.1")

        client.run("install consumer -s build_type=Debug")
        assert "dep/0.1: Main binary package '9e186f6d94c008b544af1569d1a6368d8339efc5' missing"\
               in client.out
        assert f"Found compatible package '{package_id}'" in client.out

    def test_compatible_recipe_reference(self, client):
        """ check that the recipe name can be used to filter
        """
        client.save({"pkg/conanfile.py": GenConanfile("pkg", "0.1").with_setting("build_type"),
                     "consumer/conanfile.py": GenConanfile().with_requires("pkg/0.1")})

        client.run("create pkg -s build_type=Release")

        # The compatibility doesn't fire for package "pkg"
        client.run("install consumer -s build_type=Debug", assert_error=True)
        assert "ERROR: Missing binary" in client.out


def test_cppstd():
    client = TestClient()
    compatibles = textwrap.dedent("""\
        def compatibility(conanfile):
            cppstd = conanfile.settings.get_safe("compiler.cppstd")
            if not cppstd:
                return

            result = []
            for cppstd in ["11", "14", "17", "20"]:
                result.append({"settings": [("compiler.cppstd", cppstd)]})

            if conanfile.settings.build_type == "Debug":
                for cppstd in ["11", "14", "17", "20"]:
                    result.append({"settings": [("compiler.cppstd", cppstd),
                                                ("build_type", "Release")]})
            return result
        """)
    client.save_home({"extensions/plugins/compatibility/compatibility.py": compatibles})

    conanfile = GenConanfile("dep", "0.1").with_settings("compiler", "build_type")
    client.save({"dep/conanfile.py": conanfile,
                 "consumer/conanfile.py": GenConanfile().with_requires("dep/0.1")})

    base_settings = "-s compiler=gcc -s compiler.version=7 -s compiler.libcxx=libstdc++11"
    client.run(f"create dep {base_settings} -s build_type=Release -s compiler.cppstd=14")
    package_id = client.created_package_id("dep/0.1")

    client.run(f"install consumer {base_settings} -s compiler.cppstd=17")
    assert "dep/0.1: Checking 3 compatible configurations" in client.out
    assert "dep/0.1: Main binary package 'ec174bec4a5ee2d44d3e33d9f4fdacd9b65a6772' missing" \
           in client.out
    assert f"Found compatible package '{package_id}'" in client.out

    client.run(f"install consumer {base_settings} -s build_type=Debug -s compiler.cppstd=17")
    assert "dep/0.1: Main binary package '94758b7bbcb365aaf355913b35431c0da6ed6da5' missing" \
           in client.out
    assert f"Found compatible package '{package_id}'" in client.out


def test_cppstd_validated():
    """ this test proves that 1 only configuration, the latest one, is tested and compatible,
    because the ``valiate()`` method is rejecting all cppstd<20
    """
    client = TestClient()
    compatibles = textwrap.dedent("""\
        def compatibility(conanfile):
            return [{"settings": [("compiler.cppstd", v)]} for v in ("11", "14", "17", "20")]
        """)
    client.save_home({"extensions/plugins/compatibility/compatibility.py": compatibles})

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.build import check_min_cppstd
        class Pkg(ConanFile):
            name = "dep"
            version = "0.1"
            settings = "compiler"
            def validate(self):
                check_min_cppstd(self, "20")
        """)

    client.save({"dep/conanfile.py": conanfile,
                 "consumer/conanfile.py": GenConanfile().with_requires("dep/0.1")})

    base_settings = "-s compiler=gcc -s compiler.version=8 -s compiler.libcxx=libstdc++11"
    client.run(f"create dep {base_settings} -s compiler.cppstd=20")

    client.run(f"install consumer {base_settings} -s compiler.cppstd=17", assert_error=True)
    assert "dep/0.1: Invalid: Current cppstd (17) is lower than the required C++ standard (20)." \
           in client.out


def test_cppstd_server():
    """ this test proves the order, first from cache
    """
    c = TestClient(default_server_user=True)
    compatibles = textwrap.dedent("""\
        def compatibility(conanfile):
            return [{"settings": [("compiler.cppstd", v)]} for v in ("11", "14", "17", "20")]
        """)
    c.save_home({"extensions/plugins/compatibility/compatibility.py": compatibles})

    conanfile = GenConanfile("dep", "0.1").with_settings("compiler")
    c.save({"dep/conanfile.py": conanfile,
            "consumer/conanfile.py": GenConanfile().with_requires("dep/0.1")})

    base_settings = "-s compiler=gcc -s compiler.version=8 -s compiler.libcxx=libstdc++11"
    c.run(f"create dep {base_settings} -s compiler.cppstd=20")
    c.run("upload * -r=default -c")
    c.run("remove * -c")

    c.run(f"install consumer {base_settings} -s compiler.cppstd=17")
    assert "dep/0.1: Checking 3 compatible configurations" in c.out
    assert "dep/0.1: Compatible configurations not found in cache, checking servers" in c.out
    assert "dep/0.1: Main binary package '6179018ccb6b15e6443829bf3640e25f2718b931' missing" \
           in c.out
    assert "Found compatible package '326c500588d969f55133fdda29506ef61ef03eee': " \
           "compiler.cppstd=20" in c.out
    c.assert_listed_binary({"dep/0.1": ("326c500588d969f55133fdda29506ef61ef03eee",
                                        "Download (default)")})
    # second time, not download, already in cache
    c.run(f"install consumer {base_settings} -s compiler.cppstd=17")
    assert "dep/0.1: Checking 3 compatible configurations" in c.out
    assert "dep/0.1: Compatible configurations not found in cache, checking servers" not in c.out
    assert "dep/0.1: Main binary package '6179018ccb6b15e6443829bf3640e25f2718b931' missing" in c.out
    assert "Found compatible package '326c500588d969f55133fdda29506ef61ef03eee': " \
           "compiler.cppstd=20" in c.out
    c.assert_listed_binary({"dep/0.1": ("326c500588d969f55133fdda29506ef61ef03eee", "Cache")})

    # update checks in servers
    c2 = TestClient(servers=c.servers, inputs=["admin", "password"])
    c2.save({"dep/conanfile.py": conanfile})
    c2.run(f"create dep {base_settings} -s compiler.cppstd=14")
    c2.run("upload * -r=default -c")

    c.run(f"install consumer {base_settings} -s compiler.cppstd=17 --update")
    assert "dep/0.1: Checking 3 compatible configurations" in c.out
    assert "dep/0.1: Compatible configurations not found in cache, checking servers" not in c.out
    assert "dep/0.1: Main binary package '6179018ccb6b15e6443829bf3640e25f2718b931' missing" in c.out
    assert "Found compatible package 'ce92fac7c26ace631e30875ddbb3a58a190eb601': " \
           "compiler.cppstd=14" in c.out
    c.assert_listed_binary({"dep/0.1": ("ce92fac7c26ace631e30875ddbb3a58a190eb601",
                                        "Download (default)")})

    # Now what happens if we have it in cache already
    c.run(f"install consumer {base_settings} -s compiler.cppstd=17 --update")
    assert "dep/0.1: Checking 3 compatible configurations" in c.out
    assert "dep/0.1: Main binary package '6179018ccb6b15e6443829bf3640e25f2718b931' missing" in c.out
    assert "Found compatible package 'ce92fac7c26ace631e30875ddbb3a58a190eb601': " \
           "compiler.cppstd=14" in c.out
    c.assert_listed_binary({"dep/0.1": ("ce92fac7c26ace631e30875ddbb3a58a190eb601",
                                        "Cache")})


class TestDefaultCompat:

    def test_default_cppstd_compatibility(self):
        c = TestClient()
        c.save_home({"profiles/default": ""})
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                name = "mylib"
                version = "1.0"
                package_type = "library"
                options = {"shared": [True, False]}
                default_options = {"shared": False}
                settings = "os", "arch", "compiler", "build_type"
            """)
        c.save({"conanfile.py": conanfile, "profile_build": "[settings]\nos=Windows\narch=x86_64"})
        os_ = "Windows"
        build_type = "Release"
        arch = "x86_64"
        compiler = "msvc"
        version = "191"
        cppstd = "17"
        runtime = "dynamic"
        c.run(f"create . -s os={os_} -s arch={arch} -s build_type={build_type} "
              f"-s compiler={compiler} "
              f"-s compiler.version={version} -s compiler.cppstd={cppstd} "
              f"-s compiler.runtime={runtime} -pr:b=profile_build")
        package_id1 = c.created_package_id("mylib/1.0")

        # Try to install with cppstd 14, it will find cppstd 17 as compatible
        c.run(f"install --requires=mylib/1.0@ -s os={os_} -s arch={arch} -s build_type={build_type} "
              f"-s compiler={compiler} "
              f"-s compiler.version={version} -s compiler.cppstd=14 "
              f"-s compiler.runtime={runtime} -pr:b=profile_build")
        assert "mylib/1.0: Main binary package 'e340edd75790e7156c595edebd3d98b10a2e091e' missing."\
               f"Using compatible package '{package_id1}'"

    def test_fail_with_options_deleted(self):
        """
        This test used to fail with "ConanException: option 'with_fmt_alias' doesn't exist",
        because it was removed by the package_id()
        """
        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                name = "mylib"
                version = "1.0"

                options = {"with_fmt_alias": [True, False]}
                default_options = {"with_fmt_alias": False}

                settings = "os", "arch", "compiler", "build_type"

                def package_id(self):
                    del self.info.options.with_fmt_alias

                def package_info(self):
                    self.output.warning("WITH_FMT_ALIAS={}".format(self.options.with_fmt_alias))
            """)
        c.save({"conanfile.py": conanfile})

        c.run("create . -s compiler.cppstd=14")

        c.run("create . --build=missing -s compiler.cppstd=17")
        assert "mylib/1.0: Main binary package" in c.out
        assert "Found compatible package" in c.out
        assert "Possible options are ['shared', 'header_only']" not in c.out
        assert "mylib/1.0: WARN: WITH_FMT_ALIAS=False" in c.out

    def test_header_only_build_missing(self):
        """
        this test failed with self.settings.compiler setting didn't exist
        """
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class SumConan(ConanFile):
                name = "sum"
                version = "0.1"
                settings = "os", "arch", "compiler", "build_type"
                def build(self):
                    self.output.warning("My compiler is '{}'".format(self.settings.compiler))
                def package_id(self):
                    self.info.clear()
        """)

        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create . -s compiler.cppstd=17 --build missing")
        client.assert_listed_binary({"sum/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709",
                                                 "Build")})
        # Just check that it works and doesn't fail
        assert "Installing packages" in client.out
        client.run("create . -s compiler.cppstd=14 --build missing")
        # Now it will not build, as package exist
        client.assert_listed_binary({"sum/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709",
                                                 "Cache")})
        assert "Installing packages" in client.out

    def test_check_min_cppstd(self):
        """ test that the check_min_cppstd works fine wiht compatibility, as it is based
        on ``conanfile.info.settings`` not ``conanfile.settings``
        """
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.build import check_min_cppstd, valid_min_cppstd
            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                settings = "os", "arch", "compiler", "build_type"
                def validate(self):
                    check_min_cppstd(self, "17", False)
                    self.output.info("valid standard!!")
                def package_info(self):
                    self.output.info("CPPSTD: {}".format(self.settings.compiler.cppstd))
            """)

        c = TestClient()
        c.save({"conanfile.py": conanfile})
        settings = "-s compiler=gcc -s compiler.version=9 -s compiler.libcxx=libstdc++11"
        c.run(f"create .  {settings} -s compiler.cppstd=17")
        assert "pkg/0.1: valid standard!!" in c.out
        assert "pkg/0.1: CPPSTD: 17" in c.out
        c.run(f"install {settings} --requires=pkg/0.1 -s compiler.cppstd=14", assert_error=True)
        assert "pkg/0.1: Invalid: Current cppstd (14) is lower than the required C++ standard (17)."\
               in c.out
        c.run(f"install {settings} --requires=pkg/0.1 -s compiler.cppstd=20")
        assert "valid standard!!" in c.out
        assert "pkg/0.1: CPPSTD: 17" in c.out

    def test_check_min_cstd(self):
        """ test that the check_min_cstd works fine with compatibility
        """
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.build import check_min_cstd
            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                settings = "os", "arch", "compiler", "build_type"
                languages = "C"
                def validate(self):
                    check_min_cstd(self, "17", False)
                    self.output.info("valid standard!!")
                def package_info(self):
                    self.output.info("CSTD: {}".format(self.settings.compiler.cstd))
            """)

        c = TestClient()
        c.save({"conanfile.py": conanfile})
        settings = "-s compiler=gcc -s compiler.version=14 -s compiler.libcxx=libstdc++11"
        c.run(f"create .  {settings} -s compiler.cstd=17")
        assert "pkg/0.1: valid standard!!" in c.out
        assert "pkg/0.1: CSTD: 17" in c.out
        c.run(f"install {settings} --requires=pkg/0.1 -s compiler.cstd=11", assert_error=True)
        assert "pkg/0.1: Invalid: Current cstd (11) is lower than the required C standard (17)."\
               in c.out
        c.run(f"install {settings} --requires=pkg/0.1 -s compiler.cstd=23")
        assert "valid standard!!" in c.out
        assert "pkg/0.1: CSTD: 17" in c.out

    def test_check_min_cppstd_interface(self):
        """ test that says that compatible binaries are ok, as long as the user defined
        cppstd>=14. The syntax is a bit forced, maybe we want to improve ``check_min_cppstd``
        capabilities to be able to raise ConanInvalidConfiguration too
        """
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.errors import ConanInvalidConfiguration
            from conan.tools.build import check_min_cppstd, valid_min_cppstd
            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                settings = "os", "arch", "compiler", "build_type"
                def validate(self):
                    if int(str(self.info.settings.compiler.cppstd).replace("gnu", "")) <= 14:
                        raise ConanInvalidConfiguration("incompatible cppstd!")
                    check_min_cppstd(self, "17", False)  # based on self.info
                    self.output.info("valid standard!!")
                def package_info(self):
                    self.output.info("CPPSTD: {}".format(self.settings.compiler.cppstd))
            """)

        c = TestClient()
        c.save({"conanfile.py": conanfile})
        settings = "-s compiler=gcc -s compiler.version=9 -s compiler.libcxx=libstdc++11"
        c.run(f"create .  {settings} -s compiler.cppstd=17")
        assert "pkg/0.1: valid standard!!" in c.out
        assert "pkg/0.1: CPPSTD: 17" in c.out
        c.run(f"install {settings} --requires=pkg/0.1 -s compiler.cppstd=14", assert_error=True)
        assert "valid standard!!" not in c.out
        assert "pkg/0.1: Invalid: incompatible cppstd!" in c.out
        c.run(f"install {settings} --requires=pkg/0.1 -s compiler.cppstd=20")
        assert "valid standard!!" in c.out
        assert "pkg/0.1: CPPSTD: 17" in c.out

    def test_can_create_multiple(self):
        c = TestClient()
        c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_settings("os", "arch", "compiler",
                                                                         "build_type")})
        settings = "-s os=Linux -s arch=x86_64 -s compiler=gcc -s compiler.version=9 "\
                   "-s compiler.libcxx=libstdc++11"
        c.run(f"create . {settings} -s compiler.cppstd=11")
        c.assert_listed_binary({"pkg/0.1": ("0d5f0b9d89187b4e62abb10ae409997e152db9de", "Build")})
        c.run(f"create . {settings} -s compiler.cppstd=14")
        c.assert_listed_binary({"pkg/0.1": ("145f423d315bee340546093be5b333ef5238668e", "Build")})
        c.run(f"create . {settings} -s compiler.cppstd=17")
        c.assert_listed_binary({"pkg/0.1": ("00fcbc3b6ab76a68f15e7e750e8081d57a6f5812", "Build")})

    def test_unnecessary_builds(self):
        # https://github.com/conan-io/conan/issues/15657
        c = TestClient()
        c.save({"tool/conanfile.py": GenConanfile("tool", "0.1"),
                "dep/conanfile.py": GenConanfile("dep", "0.1").with_tool_requires("tool/0.1"),
                "app/conanfile.py": GenConanfile("app", "0.1").with_requires("dep/0.1")
                                                              .with_tool_requires("tool/0.1")})
        c.run("create tool")
        c.run("create dep ")
        c.run("create app ")
        c.run("remove tool:* -c")
        c.run("install --requires=dep/0.1  --build=missing")
        assert re.search(r"Skipped binaries(\s*)tool/0.1", c.out)
        c.run("graph info --requires=app/0.1 --build=missing")
        assert re.search(r"Skipped binaries(\s*)tool/0.1", c.out)
        c.run("install --requires=app/0.1  --build=missing")
        assert re.search(r"Skipped binaries(\s*)tool/0.1", c.out)

    def test_msvc_194_fallback(self):
        c = TestClient()
        c.save_home({"profiles/default": ""})
        c.save({"conanfile.py": GenConanfile("mylib", "1.0").with_settings("os", "arch",
                                                                           "compiler", "build_type"),
                "profile_build": "[settings]\nos=Windows\narch=x86_64"})

        c.run("create . -s os=Windows -s arch=x86_64 -s build_type=Release "
              "-s compiler=msvc "
              "-s compiler.version=193 -s compiler.cppstd=17 "
              "-s compiler.runtime=dynamic -pr:b=profile_build")
        package_id1 = c.created_package_id("mylib/1.0")

        # Try to install with cppstd 14, it will find cppstd 17 as compatible
        c.run("install --requires=mylib/1.0@ -s os=Windows -s arch=x86_64 -s build_type=Release "
              "-s compiler=msvc "
              "-s compiler.version=194 -s compiler.cppstd=14 "
              "-s compiler.runtime=dynamic -pr:b=profile_build")
        assert "mylib/1.0: Main binary package 'e340edd75790e7156c595edebd3d98b10a2e091e' missing."\
               f"Using compatible package '{package_id1}'"

        c.run("install --requires=mylib/1.0@ -s os=Windows -s arch=x86_64 -s build_type=Release "
              "-s compiler=msvc "
              "-s compiler.version=194 -s compiler.cppstd=17 "
              "-s compiler.runtime=dynamic -pr:b=profile_build")
        assert "mylib/1.0: Main binary package 'e340edd75790e7156c595edebd3d98b10a2e091e' missing." \
               f"Using compatible package '{package_id1}'"


class TestErrorsCompatibility:
    """ when the plugin fails, we want a clear message and a helpful trace
    """
    def test_error_compatibility(self):
        c = TestClient()
        debug_compat = textwrap.dedent("""\
            def debug_compat(conanfile):
                other(conanfile)

            def other(conanfile):
                conanfile.settings.os
            """)
        compatibles = textwrap.dedent("""\
            from debug_compat import debug_compat
            def compatibility(conanfile):
                return debug_compat(conanfile)
            """)
        c.save_home({"extensions/plugins/compatibility/compatibility.py": compatibles,
                     "extensions/plugins/compatibility/debug_compat.py": debug_compat})

        conanfile = GenConanfile("dep", "0.1")
        c.save({"dep/conanfile.py": conanfile,
                "consumer/conanfile.py": GenConanfile().with_requires("dep/0.1")})

        c.run(f"export dep")
        c.run(f"install consumer", assert_error=True)
        assert "Error while processing 'compatibility.py' plugin for 'dep/0.1', line 3" in c.out
        assert "while calling 'debug_compat', line 2" in c.out
        assert "while calling 'other', line 5" in c.out

    def test_remove_plugin_file(self):
        c = TestClient()
        c.run("version")  # to trigger the creation
        os.remove(os.path.join(HomePaths(c.cache_folder).compatibility_plugin_path,
                               "compatibility.py"))
        c.save({"conanfile.txt": ""})
        c.run("install .", assert_error=True)
        assert "ERROR: The 'compatibility.py' plugin file doesn't exist" in c.out


def test_compatible_prev_from_cache():
    tc = TestClient()
    tc.save({"conanfile.py": GenConanfile("pkg", "0.1").with_settings("compiler")})
    settings = ("-s os=Linux -s arch=x86_64 -s compiler=gcc -s compiler.version=8 "
                "-s compiler.libcxx=libstdc++11")
    tc.run(f"create . {settings} -s=compiler.cppstd=17")
    tc.run(f"install --requires=pkg/0.1 {settings} -s=compiler.cppstd=14 -u")
    tc.assert_listed_binary({"pkg/0.1": ("6179018ccb6b15e6443829bf3640e25f2718b931",
                                         "Cache")})



import json
import os
import textwrap
import time

import pytest

from conan.api.model import RecipeReference
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


@pytest.mark.parametrize("requires", ["requires", "tool_requires"])
def test_conanfile_txt_deps_ranges(requires):
    """
    conanfile.txt locking it dependencies (with version ranges)
    """
    client = TestClient(light=True)
    client.save({"pkg/conanfile.py": GenConanfile(),
                 "consumer/conanfile.txt": f"[{requires}]\npkg/[>0.0]@user/testing"})
    client.run("create pkg --name=pkg --version=0.1 --user=user --channel=testing")
    client.run("lock create consumer/conanfile.txt")
    assert "pkg/0.1@user/testing#" in client.out

    client.run("create pkg --name=pkg --version=0.2 --user=user --channel=testing")

    client.run("install consumer/conanfile.txt")
    assert "pkg/0.1@user/testing#" in client.out
    assert "pkg/0.2" not in client.out

    os.remove(os.path.join(client.current_folder, "consumer/conan.lock"))
    client.run("install consumer/conanfile.txt")
    assert "pkg/0.2@user/testing#" in client.out
    assert "pkg/0.1" not in client.out


@pytest.mark.parametrize("command", ["install", "create", "graph info", "export-pkg"])
def test_lockfile_out(command):
    # Check that lockfile out is generated for different commands
    c = TestClient(light=True)
    c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
            "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_requires("dep/[*]")})
    c.run("create dep")
    c.run(f"{command} pkg --lockfile-out=conan.lock")
    lock = c.load("conan.lock")
    assert "dep/0.1" in lock


def test_lockfile_out_export():
    # Check that lockfile out is generated for "conan export"
    c = TestClient(light=True)
    c.save({"pkg/conanfile.py": GenConanfile("pkg", "0.1")})
    c.run("export pkg --lockfile-out=conan.lock")
    lock = c.load("conan.lock")
    assert "pkg/0.1" in lock


@pytest.mark.parametrize("requires", ["requires", "tool_requires"])
def test_conanfile_txt_deps_ranges_transitive(requires):
    """
    conanfile.txt locking it dependencies and its transitive dependencies (with version ranges)
    """
    client = TestClient(light=True)
    client.save({"dep/conanfile.py": GenConanfile(),
                 "pkg/conanfile.py": GenConanfile().with_requires("dep/[>0.0]@user/testing"),
                 "consumer/conanfile.txt": f"[{requires}]\npkg/[>0.0]@user/testing"})
    client.run("create dep --name=dep --version=0.1 --user=user --channel=testing")
    client.run("create pkg --name=pkg --version=0.1 --user=user --channel=testing")

    client.run("lock create consumer/conanfile.txt")
    assert "dep/0.1@user/testing#" in client.out
    assert "pkg/0.1@user/testing#" in client.out

    client.run("create dep --name=dep --version=0.2 --user=user --channel=testing")

    client.run("install consumer/conanfile.txt")
    assert "dep/0.1@user/testing#" in client.out
    assert "dep/0.2" not in client.out

    os.remove(os.path.join(client.current_folder, "consumer/conan.lock"))
    client.run("install consumer/conanfile.txt", assert_error=True)
    assert "dep/0.2@user/testing#" in client.out
    assert "dep/0.1" not in client.out


@pytest.mark.parametrize("requires", ["requires", "tool_requires"])
def test_conanfile_txt_strict(requires):
    """
    conanfile.txt locking it dependencies (with version ranges)
    """
    client = TestClient(light=True)
    client.save({"pkg/conanfile.py": GenConanfile(),
                 "consumer/conanfile.txt": f"[{requires}]\npkg/[>0.0]@user/testing"})
    client.run("create pkg --name=pkg --version=0.1 --user=user --channel=testing")
    client.run("lock create consumer/conanfile.txt")
    assert "pkg/0.1@user/testing#" in client.out

    client.run("create pkg --name=pkg --version=0.2 --user=user --channel=testing")
    client.run("create pkg --name=pkg --version=1.2 --user=user --channel=testing")

    # Not strict mode works
    client.save({"consumer/conanfile.txt": f"[{requires}]\npkg/[>1.0]@user/testing"})

    client.run("install consumer/conanfile.txt", assert_error=True)
    kind = "build_requires" if requires == "tool_requires" else "requires"
    assert f"Requirement 'pkg/[>1.0]@user/testing' not in lockfile '{kind}'" in client.out

    client.run("install consumer/conanfile.txt --lockfile-partial")
    assert "pkg/1.2@user/testing" in client.out
    assert "pkg/1.2" not in client.load("consumer/conan.lock")

    # test it is possible to capture new changes too, when not strict, mutating the lockfile
    client.run("install consumer/conanfile.txt --lockfile-partial --lockfile-out=conan.lock")
    assert "pkg/1.2@user/testing" in client.out
    lock = client.load("conan.lock")
    assert "pkg/1.2" in lock
    assert "pkg/0.1" in lock  # both versions are locked now
    # clean legacy versions
    client.run("lock create consumer/conanfile.txt --lockfile-out=conan.lock --lockfile-clean")
    lock = client.load("conan.lock")
    assert "pkg/1.2" in lock
    assert "pkg/0.1" not in lock


@pytest.mark.parametrize("requires", ["requires", "tool_requires"])
def test_conditional_os(requires):
    """
    conanfile.txt can lock conditional dependencies (conditional on OS for example),
    with consecutive calls to "conan lock create", augmenting the lockfile
    """
    client = TestClient(light=True)

    pkg_conanfile = textwrap.dedent(f"""
        from conan import ConanFile
        class Pkg(ConanFile):
            settings = "os"
            def requirements(self):
                if self.settings.os == "Windows":
                    self.requires("win/[>0.0]")
                else:
                    self.requires("nix/[>0.0]")
        """)
    client.save({"dep/conanfile.py": GenConanfile(),
                 "pkg/conanfile.py": pkg_conanfile,
                 "consumer/conanfile.txt": f"[{requires}]\npkg/0.1"})
    client.run("create dep --name=win --version=0.1")
    client.run("create dep --name=nix --version=0.1")

    client.run("create pkg --name=pkg --version=0.1 -s os=Windows")
    client.run("create pkg --name=pkg --version=0.1 -s os=Linux")

    client.run("lock create consumer/conanfile.txt  --lockfile-out=consumer.lock -s os=Windows"
               " -s:b os=Windows")
    assert "win/0.1#" in client.out
    assert "pkg/0.1#" in client.out
    client.run("lock create consumer/conanfile.txt  --lockfile=consumer.lock "
               "--lockfile-out=consumer.lock -s os=Linux -s:b os=Linux")
    assert "nix/0.1#" in client.out
    assert "pkg/0.1#" in client.out

    # New dependencies will not be used if using the lockfile
    client.run("create dep --name=win --version=0.2")
    client.run("create dep --name=nix --version=0.2")
    client.run("create pkg --name=pkg --version=0.1 -s os=Windows")
    client.run("create pkg --name=pkg --version=0.1 -s os=Linux")

    client.run("install consumer --lockfile=consumer.lock -s os=Windows -s:b os=Windows")
    assert "win/0.1#" in client.out
    assert "win/0.2" not in client.out
    client.run("install consumer -s os=Windows -s:b os=Windows")
    assert "win/0.2#" in client.out
    assert "win/0.1" not in client.out
    assert "nix/0.1" not in client.out

    client.run("install consumer --lockfile=consumer.lock -s os=Linux -s:b os=Linux")
    assert "nix/0.1#" in client.out
    assert "nix/0.2" not in client.out
    client.run("install consumer -s os=Linux -s:b os=Linux")
    assert "nix/0.2#" in client.out
    assert "nix/0.1" not in client.out
    assert "win/" not in client.out


@pytest.mark.parametrize("requires", ["requires", "tool_requires"])
def test_conditional_same_package(requires):
    # What happens when a conditional requires different versions of the same package?
    client = TestClient(light=True)

    pkg_conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            settings = "os"
            def requirements(self):
                if self.settings.os == "Windows":
                    self.requires("dep/0.1")
                else:
                    self.requires("dep/0.2")
        """)
    client.save({"dep/conanfile.py": GenConanfile(),
                 "pkg/conanfile.py": pkg_conanfile,
                 "consumer/conanfile.txt": f"[{requires}]\npkg/0.1"})
    client.run("create dep --name=dep --version=0.1")
    client.run("create dep --name=dep --version=0.2")

    client.run("create pkg --name=pkg --version=0.1 -s os=Windows")
    client.run("create pkg --name=pkg --version=0.1 -s os=Linux")

    client.run("lock create consumer/conanfile.txt  --lockfile-out=conan.lock -s os=Windows"
               " -s:b os=Windows")
    assert "dep/0.1#" in client.out
    assert "dep/0.2" not in client.out
    client.run("lock create consumer/conanfile.txt  --lockfile=conan.lock "
               "--lockfile-out=conan.lock -s os=Linux -s:b os=Linux")
    assert "dep/0.2#" in client.out
    assert "dep/0.1" not in client.out

    client.run("install consumer --lockfile=conan.lock --lockfile-out=win.lock -s os=Windows"
               " -s:b os=Windows")
    assert "dep/0.1#" in client.out
    assert "dep/0.2" not in client.out

    client.run("install consumer --lockfile=conan.lock --lockfile-out=linux.lock -s os=Linux"
               " -s:b os=Linux")
    assert "dep/0.2#" in client.out
    assert "dep/0.1" not in client.out


@pytest.mark.parametrize("requires", ["requires", "build_requires"])
def test_conditional_incompatible_range(requires):
    client = TestClient(light=True)

    pkg_conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            settings = "os"
            def requirements(self):
                if self.settings.os == "Windows":
                    self.requires("dep/[<1.0]")
                else:
                    self.requires("dep/[>=1.0]")
        """)
    client.save({"dep/conanfile.py": GenConanfile(),
                 "pkg/conanfile.py": pkg_conanfile,
                 "consumer/conanfile.txt": "[requires]\npkg/0.1"})
    client.run("create dep --name=dep --version=0.1")
    client.run("create dep --name=dep --version=1.1")

    client.run("create pkg --name=pkg --version=0.1 -s os=Windows")
    client.run("create pkg --name=pkg --version=0.1 -s os=Linux")

    client.run("lock create consumer/conanfile.txt  --lockfile-out=conan.lock -s os=Windows"
               " -s:b os=Windows")
    assert "dep/0.1#" in client.out
    assert "dep/1.1" not in client.out
    # The previous lock was locking dep/0.1. This new lock will not use dep/0.1 as it is outside
    # of its range, can't lock to it and will depend on dep/1.1. Both dep/0.1 for Windows and
    # dep/1.1 for Linux now coexist in the lock
    client.run("lock create consumer/conanfile.txt  --lockfile=conan.lock "
               "--lockfile-out=conan.lock -s os=Linux -s:b os=Linux")
    assert "dep/1.1#" in client.out
    assert "dep/0.1" not in client.out
    lock = client.load("conan.lock")
    assert "dep/0.1" in lock
    assert "dep/1.1" in lock

    # These will not be used, lock will avoid them
    client.run("create dep --name=dep --version=0.2")
    client.run("create dep --name=dep --version=1.2")

    client.run("install consumer --lockfile=conan.lock --lockfile-out=win.lock -s os=Windows"
               " -s:b os=Windows")
    assert "dep/0.1#" in client.out
    assert "dep/1.1" not in client.out

    client.run("install consumer --lockfile=conan.lock --lockfile-out=linux.lock -s os=Linux"
               " -s:b os=Linux")
    assert "dep/1.1#" in client.out
    assert "dep/0.1" not in client.out


@pytest.mark.parametrize("requires", ["requires", "tool_requires"])
def test_conditional_compatible_range(requires):
    client = TestClient(light=True)

    pkg_conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            settings = "os"
            def requirements(self):
                if self.settings.os == "Windows":
                    self.requires("dep/[<0.2]")
                else:
                    self.requires("dep/[>0.0]")
        """)
    client.save({"dep/conanfile.py": GenConanfile(),
                 "pkg/conanfile.py": pkg_conanfile,
                 "consumer/conanfile.txt": f"[{requires}]\npkg/0.1"})
    client.run("create dep --name=dep --version=0.1")
    client.run("create dep --name=dep --version=0.2")

    client.run("create pkg --name=pkg --version=0.1 -s os=Windows")
    client.run("create pkg --name=pkg --version=0.1 -s os=Linux")

    client.run("lock create consumer/conanfile.txt  --lockfile-out=conan.lock -s os=Linux"
               " -s:b os=Linux")
    assert "dep/0.2#" in client.out
    assert "dep/0.1" not in client.out
    client.run("lock create consumer/conanfile.txt  --lockfile=conan.lock "
               "--lockfile-out=conan.lock -s os=Windows -s:b os=Windows")
    assert "dep/0.1#" in client.out
    assert "dep/0.2" not in client.out

    # These will not be used, lock will avoid them
    client.run("create dep --name=dep --version=0.1.1")
    client.run("create dep --name=dep --version=0.3")

    client.run("install consumer --lockfile=conan.lock --lockfile-out=win.lock -s os=Windows"
               " -s:b os=Windows")
    assert "dep/0.1#" in client.out
    assert "dep/0.2" not in client.out
    assert "dep/0.1.1" not in client.out

    client.run("install consumer --lockfile=conan.lock --lockfile-out=linux.lock -s os=Linux "
               " -s:b os=Linux")
    assert "dep/0.2#" in client.out
    assert "dep/0.1" not in client.out
    assert "dep/0.3" not in client.out


def test_partial_lockfile():
    """
    make sure that a partial lockfile can be applied anywhere downstream without issues,
    as lockfiles by default are not strict
    """
    c = TestClient(light=True)
    c.save({"pkga/conanfile.py": GenConanfile("pkga"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "0.1").with_requires("pkga/[*]"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "0.1").with_requires("pkgb/[*]"),
            "app/conanfile.py": GenConanfile("app", "0.1").with_requires("pkgc/[*]")})
    c.run("create pkga --version=0.1")
    c.run("lock create pkgb --lockfile-out=b.lock")
    c.run("create pkga --version=0.2")
    c.run("create pkgb --lockfile=b.lock")
    assert "pkga/0.1" in c.out
    assert "pkga/0.2" not in c.out
    c.run("install pkgc --lockfile=b.lock --lockfile-partial")
    assert "pkga/0.1" in c.out
    assert "pkga/0.2" not in c.out
    c.run("create pkgc  --lockfile=b.lock --lockfile-partial")
    assert "pkga/0.1" in c.out
    assert "pkga/0.2" not in c.out
    c.run("create app --lockfile=b.lock --lockfile-partial")
    assert "pkga/0.1" in c.out
    assert "pkga/0.2" not in c.out
    c.run("create app --lockfile=b.lock", assert_error=True)
    assert "ERROR: Requirement 'pkgc/[*]' not in lockfile" in c.out


def test_ux_defaults():
    # Make sure the when explicit ``--lockfile`` argument, the file must exist, even if is conan.lock
    c = TestClient(light=True)
    c.save({"conanfile.txt": ""})
    c.run("install . --lockfile=conan.lock", assert_error=True)
    assert "ERROR: Lockfile doesn't exist" in c.out


class TestLockTestPackage:

    @pytest.fixture()
    def client(self):
        c = TestClient()
        test_package = textwrap.dedent("""
            from conan import ConanFile

            class TestPackageConan(ConanFile):
                settings = "os", "compiler", "arch", "build_type"
                def build_requirements(self):
                    self.tool_requires("cmake/[*]")
                def requirements(self):
                    self.requires(self.tested_reference_str)
                def test(self):
                    print("package tested")
            """)

        c.save({"cmake/conanfile.py": GenConanfile("cmake"),
                "dep/conanfile.py": GenConanfile("dep"),
                "app/conanfile.py": GenConanfile("app", "1.0").with_requires("dep/[*]"),
                "app/test_package/conanfile.py": test_package})

        c.run("create cmake --version=1.0")
        c.run("create dep --version=1.0")
        return c

    def test_lock_tool_requires_test(self, client):
        # https://github.com/conan-io/conan/issues/11763
        c = client
        with c.chdir("app"):
            c.run("lock create")
            lock = c.load("conan.lock")
            assert "cmake/1.0" not in lock
            assert "dep/1.0" in lock
            c.run("lock create test_package --lockfile=conan.lock --lockfile-out=conan.lock")
            lock = c.load("conan.lock")
            assert "cmake/1.0" in lock
            assert "dep/1.0" in lock

        c.run("create cmake --version=2.0")
        c.run("create dep --version=2.0")
        with c.chdir("app"):
            c.run("create . --lockfile=conan.lock")
            assert "cmake/1.0" in c.out
            assert "dep/1.0" in c.out
            assert "cmake/2.0" not in c.out
            assert "dep/2.0" not in c.out
            assert "package tested" in c.out

    def test_partial_approach(self, client):
        """ do not include it in the lockfile, but apply it partially, so the tool_require is
        free, will freely upgrade
        """
        c = client
        # https://github.com/conan-io/conan/issues/11763
        with c.chdir("app"):
            c.run("lock create .")
            lock = c.load("conan.lock")
            assert "cmake/1.0" not in lock
            assert "dep/1.0" in lock

        c.run("create cmake --version=2.0")
        c.run("create dep --version=2.0")
        with c.chdir("app"):
            c.run("create . --lockfile=conan.lock --lockfile-partial")
            assert "cmake/2.0" in c.out  # because it is in test_package and not locked
            assert "dep/1.0" in c.out
            assert "cmake/1.0" not in c.out
            assert "dep/2.0" not in c.out
            assert "package tested" in c.out

        # or to be more guaranteed
        with c.chdir("app"):
            c.run("create . --lockfile=conan.lock -tf=\"\"")
            assert "cmake" not in c.out
            assert "dep/1.0" in c.out
            assert "dep/2.0" not in c.out
            assert "package tested" not in c.out

            c.run("test test_package app/1.0 --lockfile=conan.lock --lockfile-partial")
            assert "cmake/1.0" not in c.out
            assert "cmake/2.0" in c.out
            assert "dep/1.0" in c.out
            assert "dep/2.0" not in c.out
            assert "package tested" in c.out

    def test_create_lock_tool_requires_test(self, client):
        """ same as above, but the full lockfile including the "test_package" can be
        obtained with a "conan test"
        """
        c = client
        with c.chdir("app"):
            c.run("create . --lockfile-out=conan.lock -tf=")
            lock = c.load("conan.lock")
            assert "cmake/1.0" not in lock
            assert "dep/1.0" in lock
            c.run("test test_package app/1.0 --lockfile-partial --lockfile=conan.lock "
                  "--lockfile-out=conan.lock")
            lock = c.load("conan.lock")
            assert "cmake/1.0" in lock
            assert "dep/1.0" in lock

        c.run("create cmake --version=2.0")
        c.run("create dep --version=2.0")
        with c.chdir("app"):
            c.run("create . --lockfile=conan.lock")
            assert "cmake/1.0" in c.out
            assert "dep/1.0" in c.out
            assert "cmake/2.0" not in c.out
            assert "dep/2.0" not in c.out
            assert "package tested" in c.out

    def test_test_package_lockfile(self):
        c = TestClient(light=True)
        test = textwrap.dedent("""
            from conan import ConanFile
            class TestBasicConanfile(ConanFile):
                def requirements(self):
                    self.requires(self.tested_reference_str)
                    self.requires("pkga/1.0")
                def test(self):
                    pass
            """)
        c.save({"pkga/conanfile.py": GenConanfile("pkga", "1.0"),
                "pkgb/conanfile.py": GenConanfile("pkgb", "1.0"),
                "pkgb/test_package/conanfile.py": test})
        c.run("create pkga")
        c.run("lock create pkgb")

        # alternative 1, relax lockfile
        c.run("create pkgb --lockfile-partial")

        # alternative 2, do not run test_package with same lockfile
        c.run("create pkgb --test-folder=")
        # the test_package can be tested later, so the lockfile-partial only affects the test_package
        c.run("test pkgb/test_package pkgb/1.0 --lockfile=pkgb/conan.lock --lockfile-partial")
        assert "Using lockfile:" in c.out

        # alternative 3, create the lockfile in the test_package
        c.run("lock create pkgb/test_package --lockfile=pkgb/conan.lock "
              "--lockfile-out=pkgb/test_package/conan.lock")
        lockfile = c.load("pkgb/test_package/conan.lock")
        assert "pkga/1.0" in lockfile
        c.run("test pkgb/test_package pkgb/1.0 --lockfile=pkgb/test_package/conan.lock "
              "--lockfile-partial")

        # alternative 4, add the test_package dependencies to the main lockfile
        c.run("lock create pkgb/test_package --lockfile=pkgb/conan.lock "
              "--lockfile-out=pkgb/conan.lock")
        lockfile = c.load("pkgb/conan.lock")
        assert "pkga/1.0" in lockfile
        c.run("create pkgb --lockfile=pkgb/conan.lock")


class TestErrorDuplicates:
    def test_error_duplicates(self):
        """ the problem is having 2 different, almost identical requires that will point to the same
        thing, with different traits and not colliding.
        Lockfiles do a ``require.ref`` update and that alters some dictionaries iteration, producing
        an infinite loop and blocking
        """
        c = TestClient(light=True)
        pkg = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                def requirements(self):
                    self.requires("dep/0.1#f8c2264d0b32a4c33f251fe2944bb642", headers=False, libs=False,
                                visible=False)
                    self.requires("dep/0.1", headers=True, libs=False, visible=False)
                """)
        c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
                "pkg/conanfile.py": pkg})
        c.run("create dep --lockfile-out=conan.lock")
        c.run("create pkg", assert_error=True)
        assert "Duplicated requirement: dep/0.1" in c.out
        c.run("create pkg --lockfile=conan.lock", assert_error=True)
        assert "Duplicated requirement: dep/0.1" in c.out

    def test_error_duplicates_reverse(self):
        """ Same as above, but order requires changed
        """
        c = TestClient(light=True)
        pkg = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                def requirements(self):
                    self.requires("dep/0.1", headers=True, libs=False, visible=False)
                    self.requires("dep/0.1#f8c2264d0b32a4c33f251fe2944bb642", headers=False, libs=False,
                                visible=False)
                """)
        c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
                "pkg/conanfile.py": pkg})
        c.run("create dep --lockfile-out=conan.lock")
        c.run("create pkg", assert_error=True)
        assert "Duplicated requirement: dep/0.1" in c.out
        c.run("create pkg --lockfile=conan.lock", assert_error=True)
        assert "Duplicated requirement: dep/0.1" in c.out

    def test_error_duplicates_revisions(self):
        """ 2 different revisions can be added without conflict, if they are not visible and not
        other conflicting traits
        """
        c = TestClient(light=True)
        pkg = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                def requirements(self):
                    self.requires("dep/0.1#f8c2264d0b32a4c33f251fe2944bb642", headers=False,
                                  libs=False, visible=False)
                    self.requires("dep/0.1#7b91e6100797b8b012eb3cdc5544800b", headers=True,
                                  libs=False, visible=False)
                """)
        c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
                "dep2/conanfile.py": GenConanfile("dep", "0.1").with_class_attribute("potato=42"),
                "pkg/conanfile.py": pkg})
        c.run("create dep --lockfile-out=conan.lock")
        c.run("create dep2 --lockfile=conan.lock --lockfile-out=conan.lock")

        c.run("create pkg")
        assert "dep/0.1#f8c2264d0b32a4c33f251fe2944bb642 - Cache" in c.out
        assert "dep/0.1#7b91e6100797b8b012eb3cdc5544800b - Cache" in c.out
        c.run("create pkg --lockfile=conan.lock")
        assert "dep/0.1#f8c2264d0b32a4c33f251fe2944bb642 - Cache" in c.out
        assert "dep/0.1#7b91e6100797b8b012eb3cdc5544800b - Cache" in c.out


def test_revision_timestamp():
    """
    https://github.com/conan-io/conan/issues/14108
    """
    c = TestClient(default_server_user=True)

    c.save({"pkg/conanfile.py": GenConanfile("pkg", "0.1")})
    # revision 0
    c.run("export pkg")
    c.save({"pkg/conanfile.py": GenConanfile("pkg", "0.1").with_class_attribute("_my=1")})
    # revision 1
    c.run("export pkg")
    rrev = c.exported_recipe_revision()
    c.run("upload *#* -r=default -c")
    c.save({"pkg/conanfile.py": GenConanfile("pkg", "0.1").with_class_attribute("_my=2")})
    # revision 2
    time.sleep(1)
    c.run("export pkg")
    latest_rrev = c.exported_recipe_revision()
    c.run("upload * -r=default -c")
    c.run("remove * -c")
    c.run("list *#* -r=default --format=json")
    list_json = json.loads(c.stdout)
    server_timestamp = list_json["default"]["pkg/0.1"]["revisions"][latest_rrev]["timestamp"]

    time.sleep(2)
    ref = RecipeReference.loads(f"pkg/0.1#{rrev}")
    # we force the lock to include the 2nd revision
    c.save({"conanfile.txt": f"[requires]\n{repr(ref)}"}, clean_first=True)

    c.run("lock create .")
    lock = c.load("conan.lock")
    lock = json.loads(lock)
    locked_ref = RecipeReference.loads(lock["requires"][0])
    assert locked_ref == ref
    assert locked_ref.timestamp and locked_ref.timestamp != server_timestamp


class TestLockfileUpdate:
    """
    Check that --update works
    """

    @pytest.mark.parametrize("requires", ["requires", "tool_requires"])
    def test_conanfile_txt_deps_ranges(self, requires):
        """
        conanfile.txt locking it dependencies (with version ranges)
        """
        c = TestClient(default_server_user=True)
        c.save({"pkg/conanfile.py": GenConanfile("pkg"),
                "consumer/conanfile.txt": f"[{requires}]\npkg/[>0.0]"})
        c.run("create pkg --version=0.1")
        c.run("create pkg --version=0.2")
        c.run("upload pkg/0.2 -r=default -c")
        c.run("remove pkg/0.2 -c")
        c.run("list *")
        assert "pkg/0.2" not in c.out
        c.run("lock create consumer/conanfile.txt --update")
        assert "pkg/0.1" not in c.out
        assert "pkg/0.2" in c.out
        lock = c.load("consumer/conan.lock")
        assert "pkg/0.1" not in lock
        assert "pkg/0.2" in lock


def test_error_test_explicit():
    # https://github.com/conan-io/conan/issues/14833
    client = TestClient(light=True)
    test = GenConanfile().with_test("pass").with_class_attribute("test_type = 'explicit'")
    client.save({"conanfile.py": GenConanfile("pkg", "0.1"),
                 "test_package/conanfile.py": test})
    client.run("lock create conanfile.py --lockfile-out=my.lock")
    client.run("create . --lockfile=my.lock")


def test_lock_error_create():
    # https://github.com/conan-io/conan/issues/15801
    c = TestClient(light=True)
    c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_package_type("build-scripts")})
    c.run("lock create . -u --lockfile-out=none.lock")
    lock = json.loads(c.load("none.lock"))
    assert lock["requires"] == []
    assert lock["build_requires"] == []
    c.run("create . --lockfile=none.lock --lockfile-out=none_updated.lock")
    # It doesn't crash, it used to
    lock = json.loads(c.load("none_updated.lock"))
    assert lock["requires"] == []
    assert len(lock["build_requires"]) == 1
    assert "pkg/0.1#4e9dba5c3041ba4c87724486afdb7eb4" in lock["build_requires"][0]


def test_lock_error():
    # https://github.com/conan-io/conan/issues/17363
    c = TestClient()
    transitive_dep = textwrap.dedent("""
        # recipes/transitive_dep/conanfile.py
        from conan import ConanFile

        class TransitiveDepConan(ConanFile):
            name = "transitive_dep"
            version = "2.0.0"
            settings = "build_type"
            """)
    build_tool = textwrap.dedent("""
        # recipes/build_tool/conanfile.py
        from conan import ConanFile

        class BuildToolConan(ConanFile):
            name = "build_tool"
            version = "1.0.0"
            settings = "build_type"

            def requirements(self):
                self.requires("transitive_dep/2.0.0")
        """)
    build_tool_test = textwrap.dedent("""
        # recipes/build_tool/test_package/conanfile.py
        from conan import ConanFile

        class TestPackageConan(ConanFile):
            settings = "build_type"
            test_type = "explicit"

            def requirements(self):
                self.requires(self.tested_reference_str)

            def test(self):
                pass
        """)
    runtime_dep = textwrap.dedent("""
        # recipes/runtime_dep/conanfile.py
        from conan import ConanFile

        class RuntimeDepConan(ConanFile):
            name = "runtime_dep"
            version = "1.2.3"
            settings = "build_type"

            def build_requirements(self):
                self.tool_requires("build_tool/1.0.0")
        """)
    consumer = textwrap.dedent("""
        # recipes/consumer/conanfile.py
        from conan import ConanFile

        class ConsumerConan(ConanFile):
            name = "consumer"
            version = "0.0.1"
            settings = "build_type"

            def requirements(self):
                self.requires("runtime_dep/1.2.3")
        """)
    c.save({"recipes/transitive_dep/conanfile.py": transitive_dep,
            "recipes/build_tool/conanfile.py": build_tool,
            "recipes/build_tool/test_package/conanfile.py": build_tool_test,
            "recipes/runtime_dep/conanfile.py": runtime_dep,
            "recipes/consumer/conanfile.py": consumer,
            })
    c.run("export recipes/transitive_dep")
    c.run("export recipes/build_tool")
    c.run("export recipes/runtime_dep")

    settings = "-s:b build_type=Release -s:h build_type=Debug"
    c.run(f"lock create recipes/consumer --no-remote {settings}")

    c.run(f"graph build-order recipes/consumer {settings} --build=missing --order-by=configuration "
          "--reduce  --format=json", redirect_stdout="build_order.json")
    assert "Using lockfile:" in c.out

    ref = "transitive_dep/2.0.0"
    c.run(f"install --tool-requires={ref} {settings} --build={ref} "
          "--lockfile=recipes/consumer/conan.lock ")
    # The test of ``trantisive_dep`` doesn't have the test_package

    ref = "build_tool/1.0.0"
    c.run(f"install --tool-requires={ref} {settings} --build={ref} "
          "--lockfile=recipes/consumer/conan.lock ")

    ref = "runtime_dep/1.2.3"
    c.run(f"install --requires={ref} {settings} --build={ref} "
          "--lockfile=recipes/consumer/conan.lock ")




import platform
import textwrap

import pytest

from conan.test.utils.tools import TestClient


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
def test_msbuildtoolchain_props_with_extra_flags():
    """
    Real test which is injecting some compiler/linker options and other dummy defines and
    checking that they are being processed somehow.

    Expected result: everything was built successfully.
    """
    profile = textwrap.dedent("""\
    include(default)

    [conf]
    tools.build:cxxflags=["/analyze:quiet"]
    tools.build:cflags+=["/doc"]
    tools.build:sharedlinkflags+=["/VERBOSE:UNUSEDLIBS"]
    tools.build:exelinkflags+=["/PDB:mypdbfile"]
    tools.build:defines+=["DEF1", "DEF2"]
    """)
    client = TestClient(path_with_spaces=False)
    client.run("new msbuild_exe -d name=hello -d version=0.1")
    client.save({"myprofile": profile})
    #  conantoolchain.props is already imported in the msbuild_exe tempalte
    client.run("create . -pr myprofile -tf=")
    assert "/analyze:quiet /doc src/hello.cpp" in client.out
    assert r"/VERBOSE:UNUSEDLIBS /PDB:mypdbfile x64\Release\hello.obj" in client.out
    assert "/D DEF1 /D DEF2" in client.out
    assert "Build succeeded." in client.out


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
def test_msbuildtoolchain_winsdk_version():
    """
    Configure sdk_version
    """
    client = TestClient(path_with_spaces=False)
    client.run("new msbuild_lib -d name=hello -d version=0.1")
    #  conantoolchain.props is already imported in the msbuild_exe tempalte
    client.run("create . -s arch=x86_64 -s compiler.version=193 "
               "-c tools.microsoft:winsdk_version=10.0")
    # I have verified also opening VS IDE that the setting is correctly configured
    # because the test always run over vcvars that already activates it
    assert "amd64 - winsdk_version=10.0 - vcvars_ver=14.3" in client.out


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
def test_msbuildtoolchain_compiler_update():
    # It only works for update=8, because 19.38 is the compiler in Github actions!
    client = TestClient(path_with_spaces=False)
    client.run("new msbuild_lib -d name=hello -d version=0.1")
    #  conantoolchain.props is already imported in the msbuild_exe tempalte
    client.run("create . -s arch=x86_64 -s compiler.version=193 -s compiler.update=8")
    # I have verified also opening VS IDE that the setting is correctly configured
    # because the test always run over vcvars that already activates it
    assert "amd64 - winsdk_version=None - vcvars_ver=14.38" in client.out
    assert "hello/0.1: _MSC_VER1938" in client.out


@pytest.mark.tool("visual_studio")
@pytest.mark.skipif(platform.system() != "Windows", reason="Only for windows")
class TestTxtCommandLineMSBuild:

    def test_declarative(self):
        conanfile = textwrap.dedent("""
            [generators]
            MSBuildToolchain
            """)
        client = TestClient()
        client.save({"conanfile.txt": conanfile})
        client.run("install .")
        self._check(client)

    @staticmethod
    def _check(client):
        assert "conanfile.txt: Generator 'MSBuildToolchain' calling 'generate()'" in client.out
        toolchain = client.load("conantoolchain.props")
        assert "<?xml version" in toolchain

    def test_command_line(self):
        client = TestClient()
        client.save({"conanfile.txt": ""})
        client.run("install . -g MSBuildToolchain")
        self._check(client)



import os
import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_layout_generate():
    """ Trying to leverage the layout knowledge at generate() time and put some files
    in the bindirs folders where executables will be built, like some dll from other
    dependencies.
    https://github.com/conan-io/conan/issues/12003
    """
    c = TestClient()
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import save, load
        class Pkg(ConanFile):
            name = "pkg"
            version = "1.0"
            def layout(self):
                self.folders.build = "mybuild"
                self.folders.generators = "mybuild/generators"
                self.cpp.build.bindirs = ["bin"]
                self.cpp.build.libdirs = ["lib"]
                self.cpp.build.includedirs = ["include"]
            def generate(self):
                f = os.path.join(self.build_folder, self.cpp.build.bindir, "myfile.txt")
                save(self, f, "mybin!!!")
                f = os.path.join(self.build_folder, self.cpp.build.libdir, "myfile.txt")
                save(self, f, "mylib!!!")
                f = os.path.join(self.build_folder, self.cpp.build.includedir, "myfile.txt")
                save(self, f, "myinclude!!!")

            def build(self):
                self.output.info(load(self, os.path.join(self.cpp.build.libdir, "myfile.txt")))
                self.output.info(load(self, os.path.join(self.cpp.build.bindir, "myfile.txt")))
                self.output.info(load(self, os.path.join(self.cpp.build.includedir, "myfile.txt")))
        """)
    c.save({"conanfile.py": conanfile})
    c.run("create . ")
    assert "pkg/1.0: mybin!!!" in c.out
    assert "pkg/1.0: mylib!!!" in c.out
    assert "pkg/1.0: myinclude!!!" in c.out
    c.run("install .")
    assert os.path.exists(os.path.join(c.current_folder, "mybuild", "bin", "myfile.txt"))
    assert os.path.exists(os.path.join(c.current_folder, "mybuild", "lib", "myfile.txt"))
    assert os.path.exists(os.path.join(c.current_folder, "mybuild", "include", "myfile.txt"))


def test_generate_source_folder():
    c = TestClient()
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout

        class Pkg(ConanFile):
            settings = "os", "compiler", "build_type", "arch"

            def layout(self):
                cmake_layout(self)

            def generate(self):
                self.output.info("PKG_CONFIG_PATH {}!".format(os.path.exists(self.source_folder)))
        """)
    c.save({"conanfile.py": conanfile})
    c.run("install .")
    assert "PKG_CONFIG_PATH True!" in c.out


def test_generate_source_folder_test_package():
    c = TestClient()
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout

        class Pkg(ConanFile):
            settings = "os", "compiler", "build_type", "arch"

            def requirements(self):
                self.requires(self.tested_reference_str)

            def layout(self):
                cmake_layout(self)

            def generate(self):
                self.output.info("PKG_CONFIG_PATH {}!".format(os.path.exists(self.source_folder)))

            def test(self):
                pass
        """)
    c.save({"conanfile.py": GenConanfile("pkg", "1.0"),
            "test_package/conanfile.py": conanfile})
    c.run("create .")
    assert "PKG_CONFIG_PATH True!" in c.out


def test_generate_build_folder_test_package():
    c = TestClient()
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout

        class Pkg(ConanFile):
            settings = "os", "compiler", "build_type", "arch"

            def requirements(self):
                self.requires(self.tested_reference_str)

            def layout(self):
                cmake_layout(self)

            def generate(self):
                self.output.info(f"build_folder in test_package: {bool(self.build_folder)}")

            def test(self):
                pass
        """)

    c.save({"conanfile.py": GenConanfile("pkg", "1.0"),
            "test_package/conanfile.py": conanfile})
    c.run("create .")
    assert f"build_folder in test_package: True" in c.out


class TestCustomTestPackage:
    def test_custom_test_package(self):
        c = TestClient(light=True)
        conanfile = GenConanfile("pkg", "0.1").with_class_attribute('test_package_folder="mytest"')
        c.save({"conanfile.py": conanfile,
                "mytest/conanfile.py": GenConanfile().with_test("self.output.info('MYTEST!')"),
                "mytest2/conanfile.py": GenConanfile().with_test("self.output.info('MYTEST2!')")})
        c.run("create .")
        assert "MYTEST!" in c.out
        c.run("create . -tf=mytest2")
        assert "MYTEST2!" in c.out

    def test_custom_test_package_subfolder(self):
        c = TestClient(light=True)
        conanfile = GenConanfile("pkg", "0.1").with_class_attribute('test_package_folder="my/test"')
        c.save({"pkg/conanfile.py": conanfile,
                "pkg/my/test/conanfile.py": GenConanfile().with_test("self.output.info('MYTEST!')")})
        c.run("create pkg")
        assert "MYTEST!" in c.out

    def test_custom_test_package_sibling(self):
        c = TestClient(light=True)
        conanfile = GenConanfile("pkg", "0.1").with_class_attribute(
            'test_package_folder="../my/test"')
        c.save({"pkg/conan/conanfile.py": conanfile,
                "pkg/my/test/conanfile.py": GenConanfile().with_test("self.output.info('MYTEST!')")})
        c.run("create pkg/conan")
        assert "MYTEST!" in c.out

    def test_custom_test_package_export_pkg(self):
        c = TestClient(light=True)
        conanfile = GenConanfile("pkg", "0.1").with_class_attribute('test_package_folder="mytest"')
        c.save({"pkg/conanfile.py": conanfile,
                "pkg/mytest/conanfile.py": GenConanfile().with_test("self.output.info('MYTEST!')")})
        c.run("export-pkg pkg")
        assert "MYTEST!" in c.out



import os
import platform

import pytest

from conan.test.utils.scm import create_local_git_repo
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient
from conan.internal.util.files import save


@pytest.mark.skipif(platform.system() == "Windows", reason="Uses symlinks")
def test_custom_symlinked_home_config_install():
    base_cache = temp_folder()
    real_cache = os.path.join(base_cache, "real_cache")
    os.makedirs(real_cache)
    symlink_cache = os.path.join(base_cache, "symlink_cache")
    os.symlink(real_cache, symlink_cache)
    origin_folder = temp_folder()
    save(os.path.join(origin_folder, "myfile.txt"), "some contents")
    create_local_git_repo(folder=origin_folder)
    c = TestClient(cache_folder=symlink_cache)
    c.run(f'config install "{origin_folder}" --type=git')
    assert "Copying file myfile.txt to" in c.out



import os
import pytest

from conan.errors import ConanException
from conans.server.conf import ConanServerConfigParser
from conan.test.utils.test_files import temp_folder
from conan.internal.util.files import mkdir, save


class TestServerConfigParse:

    def test_not_allowed_encoding_password(self):
        tmp_dir = temp_folder()
        server_conf = """
[server]
jwt_secret: 534534534
jwt_expire_minutes: 120
ssl_enabled: False
port: 9300
public_port:
host_name: localhost
store_adapter: disk
authorize_timeout: 1800
disk_storage_path: ~/.conan_server/data
disk_authorize_timeout: 1800
updown_secret: tbsiGzeEygYSCcNrSYcuzmZr


[write_permissions]

[users]
demo: %s
        """
        server_dir = os.path.join(tmp_dir, ".conan_server")
        mkdir(server_dir)
        conf_path = os.path.join(server_dir, "server.conf")

        save(conf_path, server_conf % "cönan")

        server_config = ConanServerConfigParser(tmp_dir)
        with pytest.raises(ConanException, match="Password contains invalid characters. Only ASCII encoding is supported"):
            server_config.users

        save(conf_path, server_conf % "manol ito!@")
        server_config = ConanServerConfigParser(tmp_dir)
        assert server_config.users == {"demo": "manol ito!@"}

        # Now test from ENV
        server_config = ConanServerConfigParser(tmp_dir, environment={"CONAN_SERVER_USERS": "demo: cönan"})
        with pytest.raises(ConanException, match="Password contains invalid characters. Only ASCII encoding is supported"):
            server_config.users

        server_config = ConanServerConfigParser(tmp_dir, environment={"CONAN_SERVER_USERS": "demo:manolito!@"})
        assert server_config.users == {"demo": "manolito!@"}

    def test_relative_public_url(self):
        tmp_dir = temp_folder()
        server_conf = """
[server]

[write_permissions]

[users]
        """
        server_dir = os.path.join(tmp_dir, ".conan_server")
        mkdir(server_dir)
        conf_path = os.path.join(server_dir, "server.conf")
        save(conf_path, server_conf)

        server_config = ConanServerConfigParser(tmp_dir)
        assert server_config.public_url == "v2"

    def test_custom_server_folder_path(self):
        tmp_dir = temp_folder()
        server_dir = os.path.join(tmp_dir, ".custom_conan_server")
        mkdir(server_dir)
        conf_path = os.path.join(server_dir, "server.conf")
        server_conf = """
[server]

[write_permissions]

[users]
        """
        save(conf_path, server_conf)
        server_config = ConanServerConfigParser(server_dir, is_custom_path=True)
        assert server_config.conan_folder == server_dir

    def test_custom_server_path_has_custom_data_path(self):
        tmp_dir = temp_folder()
        server_dir = os.path.join(tmp_dir, ".custom_conan_server")
        mkdir(server_dir)
        conf_path = os.path.join(server_dir, "server.conf")
        server_conf = """
[server]
disk_storage_path: ./custom_data

[write_permissions]

[users]
        """
        save(conf_path, server_conf)
        server_config = ConanServerConfigParser(server_dir, is_custom_path=True)
        assert server_config.disk_storage_path == os.path.join(server_dir, "custom_data")



import os

from requests import ConnectionError

from conan.internal.paths import CONAN_MANIFEST
from conan.test.utils.tools import TestClient, TestRequester, TestServer, GenConanfile
from conan.test.utils.env import environment_update


class BadConnectionUploader(TestRequester):
    fail_on = 1

    def __init__(self, *args, **kwargs):
        super(BadConnectionUploader, self).__init__(*args, **kwargs)
        self.counter_fail = 0

    def put(self, *args, **kwargs):
        self.counter_fail += 1
        if self.counter_fail == self.fail_on:
            raise ConnectionError("Can't connect because of the evil mock")
        else:
            return super(BadConnectionUploader, self).put(*args, **kwargs)


class TerribleConnectionUploader(BadConnectionUploader):
    def put(self, *args, **kwargs):
        raise ConnectionError("Can't connect because of the evil mock")


class FailPairFilesUploader(BadConnectionUploader):

    def put(self, *args, **kwargs):
        self.counter_fail += 1
        if self.counter_fail % 2 == 1:
            raise ConnectionError("Pair file, error!")
        else:
            return super(BadConnectionUploader, self).put(*args, **kwargs)


def test_try_upload_bad_recipe():
    client = TestClient(default_server_user=True, light=True)
    client.save({"conanfile.py": GenConanfile("hello0", "1.2.1")})
    client.run("export . --user=frodo --channel=stable")
    layout = client.exported_layout()
    os.unlink(os.path.join(layout.export(), CONAN_MANIFEST))
    client.run("upload %s -r default" % str(layout.reference), assert_error=True)
    assert "Cannot upload corrupted recipe" in client.out


def test_upload_with_pattern():
    client = TestClient(default_server_user=True, light=True)
    for num in range(3):
        client.save({"conanfile.py": GenConanfile("hello{}".format(num), "1.2.1")})
        client.run("export . --user=frodo --channel=stable")

    client.run("upload hello* --confirm -r default")
    for num in range(3):
        assert "Uploading recipe 'hello%s/1.2.1@frodo/stable" % num in client.out

    client.run("upload hello0* --confirm -r default")
    assert f"'hello0/1.2.1@frodo/stable#761f54e34d59deb172d6078add7050a7' "\
           "already in server, skipping upload" in client.out
    assert "hello1" not in client.out
    assert "hello2" not in client.out


def test_check_upload_confirm_question():
    server = TestServer()
    client = TestClient(servers={"default": server},
                        inputs=["yes", "admin", "password", "n", "n", "n"], light=True)
    client.save({"conanfile.py": GenConanfile("hello1", "1.2.1")})
    client.run("export . --user=frodo --channel=stable")
    client.run("upload hello* -r default")

    assert "Uploading recipe 'hello1/1.2.1@frodo/stable" in client.out

    client.save({"conanfile.py": GenConanfile("hello2", "1.2.1")})
    client.run("create . --user=frodo --channel=stable")
    client.run("upload hello* -r default")

    assert "Uploading recipe 'hello2/1.2.1@frodo/stable" not in client.out


def test_check_upload_confirm_question_yes():
    server = TestServer()
    client = TestClient(servers={"default": server},
                        inputs=["yes", "yes", "yes", "yes", "yes", "admin", "password"],
                        light=True)
    client.save({"conanfile.py": GenConanfile("hello1", "1.2.1")})
    client.run("create . ")
    client.save({"conanfile.py": GenConanfile("hello1", "1.2.1").with_package_file("file.txt",
                                                                                   env_var="MYVAR")})

    with environment_update({"MYVAR": "0"}):
        client.run("create . ")
    with environment_update({"MYVAR": "1"}):
        client.run("create . ")
    client.run("upload hello*#*:*#* -r default")
    assert str(client.out).count("(Uploaded)") == 5


class TestUpload:

    def _get_client(self, requester=None):
        servers = {}
        # All can write (for avoid authentication until we mock user_io)
        self.test_server = TestServer([("*/*@*/*", "*")], [("*/*@*/*", "*")],
                                      users={"lasote": "mypass"})
        servers["default"] = self.test_server
        test_client = TestClient(servers=servers, inputs=["lasote", "mypass"],
                                 requester_class=requester, light=True)
        return test_client

    def test_upload_error(self):
        """Cause an error in the transfer and see some message"""

        # Check for the default behaviour
        client = self._get_client(BadConnectionUploader)
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1").with_exports("*")}
        client.save(files)
        client.run("export . --user=frodo --channel=stable")
        client.run("upload hello* --confirm -r default")
        assert "Can't connect because of the evil mock" in client.out
        assert "Waiting 5 seconds to retry..." in client.out

        # This will fail in the first put file, so, as we need to
        # upload 3 files (conanmanifest, conanfile and tgz) will do it with 2 retries
        client = self._get_client(BadConnectionUploader)
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1").with_exports("*")}
        client.save(files)
        client.run("export . --user=frodo --channel=stable")
        self._set_global_conf(client, retry_wait=0)
        client.run("upload hello* --confirm -r default")
        assert "Can't connect because of the evil mock" in client.out
        assert "Waiting 0 seconds to retry..." in client.out

        # but not with 0
        client = self._get_client(BadConnectionUploader)
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1").with_exports("*"),
                 "somefile.txt": ""}
        client.save(files)
        client.run("export . --user=frodo --channel=stable")
        self._set_global_conf(client, retry=0, retry_wait=1)
        client.run("upload hello* --confirm -r default", assert_error=True)
        assert "Waiting 1 seconds to retry..." not in client.out
        assert ("Execute upload again to retry upload the failed files: "
                "conan_export.tgz. [Remote: default]") in client.out

        # Try with broken connection even with 10 retries
        client = self._get_client(TerribleConnectionUploader)
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1").with_exports("*")}
        client.save(files)
        client.run("export . --user=frodo --channel=stable")
        self._set_global_conf(client, retry=10, retry_wait=0)
        client.run("upload hello* --confirm -r default", assert_error=True)
        assert "Waiting 0 seconds to retry..." in client.out
        assert "Execute upload again to retry upload the failed files" in client.out

        # For each file will fail the first time and will success in the second one
        client = self._get_client(FailPairFilesUploader)
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1").with_exports("*")}
        client.save(files)
        client.run("export . --user=frodo --channel=stable")
        client.run("install --requires=hello0/1.2.1@frodo/stable --build='*' -r default")
        self._set_global_conf(client, retry=3, retry_wait=0)
        client.run("upload hello* --confirm -r default")
        assert str(client.out).count("WARN: network: Pair file, error!") == 5

    @staticmethod
    def _set_global_conf(client, retry=None, retry_wait=None):
        lines = []
        if retry is not None:
            lines.append("core.upload:retry={}".format(retry))
        if retry_wait is not None:
            lines.append("core.upload:retry_wait={}".format(retry_wait))

        client.save_home({"global.conf": "\n".join(lines)})

    def test_upload_error_with_config(self):
        """Cause an error in the transfer and see some message"""

        # This will fail in the first put file, so, as we need to
        # upload 3 files (conanmanifest, conanfile and tgz) will do it with 2 retries
        client = self._get_client(BadConnectionUploader)
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1").with_exports("*")}
        client.save(files)
        client.run("export . --user=frodo --channel=stable")
        self._set_global_conf(client, retry_wait=0)

        client.run("upload hello* --confirm -r default")
        assert "Can't connect because of the evil mock" in client.out
        assert "Waiting 0 seconds to retry..." in client.out

        # but not with 0
        client = self._get_client(BadConnectionUploader)
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1").with_exports("*"),
                 "somefile.txt": ""}
        client.save(files)
        client.run("export . --user=frodo --channel=stable")

        self._set_global_conf(client, retry=0, retry_wait=1)
        client.run("upload hello* --confirm -r default", assert_error=True)
        assert "Waiting 1 seconds to retry..." not in client.out
        assert ("Execute upload again to retry upload the failed files: "
                "conan_export.tgz. [Remote: default]") in client.out

        # Try with broken connection even with 10 retries
        client = self._get_client(TerribleConnectionUploader)
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1").with_exports("*")}
        client.save(files)
        client.run("export . --user=frodo --channel=stable")
        self._set_global_conf(client, retry=10, retry_wait=0)
        client.run("upload hello* --confirm -r default", assert_error=True)
        assert "Waiting 0 seconds to retry..." in client.out
        assert "Execute upload again to retry upload the failed files" in client.out

        # For each file will fail the first time and will success in the second one
        client = self._get_client(FailPairFilesUploader)
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1").with_exports("*")}
        client.save(files)
        client.run("export . --user=frodo --channel=stable")
        client.run("install --requires=hello0/1.2.1@frodo/stable --build='*'")
        self._set_global_conf(client, retry=3, retry_wait=0)
        client.run("upload hello* --confirm -r default")
        assert str(client.out).count("WARN: network: Pair file, error!") == 5

    def test_upload_same_package_dont_compress(self):
        client = self._get_client()
        client.save({"conanfile.py": GenConanfile().with_exports_sources("*"), "content.txt": "foo"})
        client.run("create . --name foo --version 1.0")

        client.run("upload foo/1.0 -r default")
        assert "foo/1.0: Compressing conan_sources.tgz" in client.out
        assert ("foo/1.0:da39a3ee5e6b4b0d3255bfef95601890afd80709: "
                "Compressing conan_package.tgz") in client.out

        client.run("upload foo/1.0 -r default")
        assert "Compressing" not in client.out
        assert "already in server, skipping upload" in client.out



import os
import textwrap
import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient, NO_SETTINGS_PACKAGE_ID
from conan.internal.util.files import mkdir


class TestSetVersionName:

    @pytest.mark.parametrize("user_channel", ["", "@user/channel"])
    def test_set_version_name(self, user_channel):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Lib(ConanFile):
                def set_name(self):
                    self.name = "pkg"
                def set_version(self):
                    self.version = "2.1"
            """)
        client.save({"conanfile.py": conanfile})
        user_channel_arg = "--user=user --channel=channel" if user_channel else ""
        client.run("export . %s" % user_channel_arg)
        assert "pkg/2.1%s: Exported" % user_channel in client.out
        # installing it doesn't break
        client.run("install --requires=pkg/2.1%s --build=missing" % (user_channel or "@"))
        client.assert_listed_require({f"pkg/2.1{user_channel}": "Cache"})
        client.assert_listed_binary({f"pkg/2.1{user_channel}": (NO_SETTINGS_PACKAGE_ID, "Build")})

        client.run("install --requires=pkg/2.1%s --build=missing" % (user_channel or "@"))
        client.assert_listed_require({f"pkg/2.1{user_channel}": "Cache"})
        client.assert_listed_binary({f"pkg/2.1{user_channel}": (NO_SETTINGS_PACKAGE_ID, "Cache")})

        # Local flow should also work
        client.run("install .")
        assert "conanfile.py (pkg/2.1):" in client.out

    def test_set_version_name_file(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import load
            class Lib(ConanFile):
                def set_name(self):
                    self.name = load(self, "name.txt")
                def set_version(self):
                    self.version = load(self, "version.txt")
            """)
        client.save({"conanfile.py": conanfile,
                     "name.txt": "pkg",
                     "version.txt": "2.1"})
        client.run("export . --user=user --channel=testing")
        assert "pkg/2.1@user/testing: Exported" in client.out
        client.run("install --requires=pkg/2.1@user/testing --build=missing")
        client.assert_listed_binary({f"pkg/2.1@user/testing": (NO_SETTINGS_PACKAGE_ID, "Build")})
        client.run("install --requires=pkg/2.1@user/testing")
        client.assert_listed_binary({f"pkg/2.1@user/testing": (NO_SETTINGS_PACKAGE_ID, "Cache")})
        # Local flow should also work
        client.run("install .")
        assert "conanfile.py (pkg/2.1):" in client.out

    def test_set_version_name_errors(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Lib(ConanFile):
                def set_name(self):
                    self.name = "pkg"
                def set_version(self):
                    self.version = "2.1"
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=other --version=1.1 --user=user --channel=testing")
        assert "pkg/2.1@user/testing" in client.out
        client.run("export .  --version=1.1 --user=user --channel=testing")
        assert "pkg/2.1@user/testing" in client.out
        # These are checked but match and don't conflict
        client.run("export . --version=2.1 --user=user --channel=testing")
        client.run("export . --name=pkg --version=2.1 --user=user --channel=testing")

        # Local flow should also fail
        client.run("install . --name=other --version=1.2")
        assert "" in client.out

    def test_set_version_name_appending(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
               from conan import ConanFile
               class Lib(ConanFile):
                   def set_name(self):
                       self.name = self.name + ".extra"
                   def set_version(self):
                       self.version = self.version + ".extra"
               """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=pkg --version=1.0")
        assert "pkg.extra/1.0.extra: Exported" in client.out

    def test_set_version_name_only_not_cli(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Lib(ConanFile):
                def set_name(self):
                    self.name = self.name or "pkg"
                def set_version(self):
                    self.version = self.version or "2.0"
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=other --version=1.1 --user=user --channel=testing")
        assert "other/1.1@user/testing: Exported" in client.out
        client.run("export .  --version=1.1 --user=user --channel=testing")
        assert "pkg/1.1@user/testing: Exported" in client.out
        client.run("export . --user=user --channel=testing")
        assert "pkg/2.0@user/testing: Exported" in client.out

        # Local flow should also work
        client.run("install . --name=other --version=1.2")
        assert "conanfile.py (other/1.2)" in client.out
        client.run("install .")
        assert "conanfile.py (pkg/2.0)" in client.out

    def test_set_version_name_crash(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Lib(ConanFile):
                def set_name(self):
                    self.name = error
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export .", assert_error=True)
        assert "ERROR: conanfile.py: Error in set_name() method, line 5" in client.out
        assert "name 'error' is not defined" in client.out
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           class Lib(ConanFile):
               def set_version(self):
                   self.version = error
           """)
        client.save({"conanfile.py": conanfile})
        client.run("export .", assert_error=True)
        assert "ERROR: conanfile.py: Error in set_version() method, line 5" in client.out
        assert "name 'error' is not defined" in client.out

    def test_set_version_cwd(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.files import load
            class Lib(ConanFile):
                name = "pkg"
                def set_version(self):
                    self.version = load(self, "version.txt")
            """)
        client.save({"conanfile.py": conanfile})
        mkdir(os.path.join(client.current_folder, "build"))
        with client.chdir("build"):
            client.save({"version.txt": "2.1"}, clean_first=True)
            client.run("export .. ")
            assert "pkg/2.1: Exported" in client.out


def test_set_version_forbidden_chars():
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("pkg", "$$$")})
    c.run("export .", assert_error=True)
    assert "ERROR: Invalid package version '$$$'" in c.out

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Lib(ConanFile):
            name = "pkg"
            def set_version(self):
                self.version = "$$$"
        """)
    c.save({"conanfile.py": conanfile})
    c.run("export .", assert_error=True)
    assert "ERROR: Invalid package version '$$$'" in c.out


def test_set_version_carriage_return():
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("pkg", r"1.0\n")})
    c.run("export .", assert_error=True)
    assert "ERROR: Invalid package version '1.0" in c.out

    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import load
        class Lib(ConanFile):
            name = "pkg"
            def set_version(self):
                self.version = load(self, "version.txt")
        """)
    c.save({"conanfile.py": conanfile,
            "version.txt": "1.0\n"})
    c.run("export .", assert_error=True)
    assert "ERROR: Invalid package version '1.0" in c.out


def test_set_version_channel():
    """
    Dynamically setting the user and channel is possible with the ``set_version()``
    not recommended, but possible.
    """
    c = TestClient()

    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import load
        class Lib(ConanFile):
            name = "pkg"
            version = "0.1"
            def set_version(self):
                self.user = "myuser"
                self.channel = "mychannel"
        """)
    c.save({"conanfile.py": conanfile})
    c.run("export .")
    assert "pkg/0.1@myuser/mychannel: Exported" in c.out



import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_cpp_info_editable():

    client = TestClient()

    conan_hello = str(GenConanfile())
    conan_hello += """
    def layout(self):
        self.folders.source = "my_sources"
        self.folders.build = "my_build"

        self.cpp.build.includedirs = ["my_include"]
        self.cpp.build.libdirs = ["my_libdir"]
        self.cpp.build.libs = ["hello"]
        self.cpp.build.objects = ["myobjs/myobject.o"]
        self.cpp.build.frameworkdirs = []  # Empty list is also explicit priority declaration

        self.cpp.source.cxxflags = ["my_cxx_flag"]
        self.cpp.source.includedirs = ["my_include_source"]
        self.cpp.source.builddirs = ["my_builddir_source"]
        self.cpp.source.set_property("cmake_build_modules", ["mypath/mybuildmodule"])

        self.cpp.package.libs = ["lib_when_package"]

    def package_info(self):
        # when editable: This one will be discarded because declared in build
        self.cpp_info.includedirs = ["package_include"]

        # when editable: This one will be discarded because declared in build
        self.cpp_info.libs.append("lib_when_package2")

        self.cpp_info.objects = ["myobject.o"]
        self.cpp_info.set_property("cmake_build_modules", ["mymodules/mybuildmodule"])

        # when editable: This one will be discarded because declared in source
        self.cpp_info.cxxflags.append("my_cxx_flag2")

        # when editable: This one will be discarded because declared in source
        self.cpp_info.frameworkdirs.append("package_frameworks_path")

        # when editable: This one WONT be discarded as has not been declared in the editables layout
        self.cpp_info.cflags.append("my_c_flag")
     """

    client.save({"conanfile.py": conan_hello})
    client.run("create . --name=hello --version=1.0")
    package_folder = client.created_layout().package().replace("\\", "/") + "/"

    conan_consumer = textwrap.dedent("""
    import os

    from conan import ConanFile, tools

    class HelloTestConan(ConanFile):
        settings = "os", "compiler", "build_type", "arch"
        requires = "hello/1.0"
        def build(self):
            info = self.dependencies["hello"].cpp_info

            self.output.warning("**includedirs:{}**".format(info.includedirs))
            self.output.warning("**libdirs:{}**".format(info.libdirs))
            self.output.warning("**builddirs:{}**".format(info.builddirs))
            self.output.warning("**frameworkdirs:{}**".format(info.frameworkdirs))
            self.output.warning("**libs:{}**".format(info.libs))
            self.output.warning("**objects:{}**".format(info.objects))
            self.output.warning("**build_modules:{}**".format(info.get_property("cmake_build_modules")))
            self.output.warning("**cxxflags:{}**".format(info.cxxflags))
            self.output.warning("**cflags:{}**".format(info.cflags))
    """)
    # When hello is not editable
    client2 = TestClient(client.cache_folder)
    client2.save({"conanfile.py": conan_consumer})
    client2.run("create . --name=lib --version=1.0")
    out = str(client2.out).replace(r"\\", "/").replace(package_folder, "")
    assert "**includedirs:['package_include']**" in out
    assert "**libdirs:['lib']**" in out
    assert "**builddirs:[]**" in out
    assert "**frameworkdirs:['package_frameworks_path']**" in out
    assert "**libs:['lib_when_package', 'lib_when_package2']**" in out
    assert "**objects:['myobject.o']**" in out
    assert "**build_modules:['mymodules/mybuildmodule']**" in out
    assert "**cxxflags:['my_cxx_flag2']**" in out
    assert "**cflags:['my_c_flag']**" in out

    # When hello is editable
    client.save({"conanfile.py": conan_hello})
    client.run("editable add . --name=hello --version=1.0")

    # Create the consumer again, now it will use the hello editable
    client2.run("create . --name=lib --version=1.0")
    base_folder = client.current_folder.replace("\\", "/") + "/"
    out = str(client2.out).replace(r"\\", "/").replace(base_folder, "")

    assert "**includedirs:['my_sources/my_include_source', 'my_build/my_include']**" in out
    assert "**libdirs:['my_build/my_libdir']**" in out
    assert "**builddirs:['my_sources/my_builddir_source']**" in out
    assert "**libs:['hello']**" in out
    assert "**objects:['my_build/myobjs/myobject.o']**" in out
    assert "**build_modules:['my_sources/mypath/mybuildmodule']**" in out
    assert "**cxxflags:['my_cxx_flag']**" in out
    assert "**cflags:['my_c_flag']**" in out
    assert "**frameworkdirs:[]**" in out


def test_cpp_info_components_editable():

    client = TestClient()

    conan_hello = str(GenConanfile())
    conan_hello += """
    def layout(self):
        self.folders.source = "my_sources"
        self.folders.build = "my_build"

        self.cpp.build.components["foo"].includedirs = ["my_include_foo"]
        self.cpp.build.components["foo"].libdirs = ["my_libdir_foo"]
        self.cpp.build.components["foo"].libs = ["hello_foo"]
        self.cpp.build.components["foo"].objects = ["myobjs/myobject.o"]

        self.cpp.build.components["var"].includedirs = ["my_include_var"]
        self.cpp.build.components["var"].libdirs = ["my_libdir_var"]
        self.cpp.build.components["var"].libs = ["hello_var"]

        self.cpp.source.components["foo"].cxxflags = ["my_cxx_flag_foo"]
        self.cpp.source.components["foo"].includedirs = ["my_include_source_foo"]
        self.cpp.source.components["foo"].builddirs = ["my_builddir_source_foo"]
        self.cpp.source.components["foo"].set_property("cmake_build_modules",
                                                      ["mypath/mybuildmodule"])
        self.cpp.source.components["var"].cxxflags = ["my_cxx_flag_var"]
        self.cpp.source.components["var"].includedirs = ["my_include_source_var"]
        self.cpp.source.components["var"].builddirs = ["my_builddir_source_var"]

        self.cpp.package.components["foo"].libs = ["lib_when_package_foo"]
        self.cpp.package.components["var"].libs = ["lib_when_package_var"]

    def package_info(self):
        # when editable: This one will be discarded because declared in build
        self.cpp_info.components["foo"].includedirs = ["package_include_foo"]

        # when editable: This one will be discarded because declared in build
        self.cpp_info.components["foo"].libs.append("lib_when_package2_foo")

        self.cpp_info.components["foo"].objects = ["myobject.o"]
        self.cpp_info.components["foo"].set_property("cmake_build_modules",
                                                    ["mymodules/mybuildmodule"])

        # when editable: This one will be discarded because declared in source
        self.cpp_info.components["foo"].cxxflags.append("my_cxx_flag2_foo")

        # when editable: This one WONT be discarded as has not been declared in the editables layout
        self.cpp_info.components["foo"].cflags.append("my_c_flag_foo")

        # when editable: This one will be discarded because declared in build
        self.cpp_info.components["var"].includedirs = ["package_include_var"]

        # when editable: This one will be discarded because declared in build
        self.cpp_info.components["var"].libs.append("lib_when_package2_var")

        # when editable: This one will be discarded because declared in source
        self.cpp_info.components["var"].cxxflags.append("my_cxx_flag2_var")

        # when editable: This one WONT be discarded as has not been declared in the editables layout
        self.cpp_info.components["var"].cflags.append("my_c_flag_var")
     """

    client.save({"conanfile.py": conan_hello})
    client.run("create . --name=hello --version=1.0")
    package_folder = client.created_layout().package().replace("\\", "/") + "/"

    conan_consumer = textwrap.dedent("""
    import os

    from conan import ConanFile, tools

    class HelloTestConan(ConanFile):
        settings = "os", "compiler", "build_type", "arch"
        requires = "hello/1.0"
        def build(self):
            info = self.dependencies["hello"].cpp_info

            self.output.warning("**FOO includedirs:{}**".format(info.components["foo"].includedirs))
            self.output.warning("**FOO libdirs:{}**".format(info.components["foo"].libdirs))
            self.output.warning("**FOO builddirs:{}**".format(info.components["foo"].builddirs))
            self.output.warning("**FOO libs:{}**".format(info.components["foo"].libs))
            self.output.warning("**FOO cxxflags:{}**".format(info.components["foo"].cxxflags))
            self.output.warning("**FOO cflags:{}**".format(info.components["foo"].cflags))
            self.output.warning("**FOO objects:{}**".format(info.components["foo"].objects))
            self.output.warning("**FOO build_modules:{}**".format(
                                        info.components["foo"].get_property("cmake_build_modules")))

            self.output.warning("**VAR includedirs:{}**".format(info.components["var"].includedirs))
            self.output.warning("**VAR libdirs:{}**".format(info.components["var"].libdirs))
            self.output.warning("**VAR builddirs:{}**".format(info.components["var"].builddirs))
            self.output.warning("**VAR libs:{}**".format(info.components["var"].libs))
            self.output.warning("**VAR cxxflags:{}**".format(info.components["var"].cxxflags))
            self.output.warning("**VAR cflags:{}**".format(info.components["var"].cflags))
    """)
    # When hello is not editable
    client2 = TestClient(client.cache_folder)
    client2.save({"conanfile.py": conan_consumer})
    client2.run("create . --name=lib --version=1.0")

    out = str(client2.out).replace(r"\\", "/").replace(package_folder, "")
    assert "**FOO includedirs:['package_include_foo']**" in out
    assert "**FOO libdirs:['lib']**" in out  # The components does have default dirs
    assert "**FOO builddirs:[]**" in out  # The components don't have default dirs for builddirs

    assert "**FOO libs:['lib_when_package_foo', 'lib_when_package2_foo']**" in out
    assert "**FOO objects:['myobject.o']**" in out
    assert "**FOO build_modules:['mymodules/mybuildmodule']**" in out
    assert "**FOO cxxflags:['my_cxx_flag2_foo']**" in out
    assert "**FOO cflags:['my_c_flag_foo']**" in out

    assert "**VAR includedirs:['package_include_var']**" in out
    assert "**VAR libdirs:['lib']**" in out  # The components does have default dirs
    assert "**VAR builddirs:[]**" in out  # The components don't have default dirs

    assert "**VAR libs:['lib_when_package_var', 'lib_when_package2_var']**" in out
    assert "**VAR cxxflags:['my_cxx_flag2_var']**" in out
    assert "**VAR cflags:['my_c_flag_var']**" in out

    # When hello is editable
    client.save({"conanfile.py": conan_hello})
    client.run("editable add . --name=hello --version=1.0")

    # Create the consumer again, now it will use the hello editable
    client2.run("create . --name=lib --version=1.0")
    base_folder = client.current_folder.replace("\\", "/") + "/"
    out = str(client2.out).replace(r"\\", "/").replace(base_folder, "")

    assert "**FOO includedirs:['my_sources/my_include_source_foo', " \
           "'my_build/my_include_foo']**" in out
    assert "**FOO libdirs:['my_build/my_libdir_foo']**" in out
    assert "**FOO builddirs:['my_sources/my_builddir_source_foo']**" in out
    assert "**FOO libs:['hello_foo']**" in out
    assert "**FOO objects:['my_build/myobjs/myobject.o']**" in out
    assert "**FOO build_modules:['my_sources/mypath/mybuildmodule']**" in out
    assert "**FOO cxxflags:['my_cxx_flag_foo']**" in out
    assert "**FOO cflags:['my_c_flag_foo']**" in out

    assert "**VAR includedirs:['my_sources/my_include_source_var', " \
           "'my_build/my_include_var']**" in out
    assert "**VAR libdirs:['my_build/my_libdir_var']**" in out
    assert "**VAR builddirs:['my_sources/my_builddir_source_var']**" in out
    assert "**VAR libs:['hello_var']**" in out
    assert "**VAR cxxflags:['my_cxx_flag_var']**" in out
    assert "**VAR cflags:['my_c_flag_var']**" in out


def test_editable_package_folder():
    """ This test checks the behavior that self.package_folder is NOT defined (i.e = None)
    for editable packages, so it cannot be used in ``package_info()`` method
    """
    c = TestClient()
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout
        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "os", "compiler", "arch", "build_type"

            def package_info(self):
                self.output.info("PKG FOLDER={}!!!".format(self.package_folder))

            def layout(self):
                cmake_layout(self)
        """)
    c.save({"conanfile.py": conanfile})
    c.run("create .")
    c.run("editable add .")
    c.run("install --requires=pkg/0.1@")
    assert "pkg/0.1: PKG FOLDER=None!!!"


def test_editable_components_absolute_paths():
    """
    this was failing in https://github.com/conan-io/conan/issues/14777
    because components aggregation had a bug
    """
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):
            name = "alpha"
            version = "1.0"

            def layout(self):
                self.cpp.source.components["headers"].includedirs = ["include222"]
                self.cpp.build.components["alpha1"].libdirs = ["alpha1",]
                self.cpp.build.components["alpha2"].libdirs = ["alpha2"]
                self.cpp.build.components["alpha_exe"].bindirs = ["alpha_exe"]
        """)
    test = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeDeps"

            def requirements(self):
                self.requires(self.tested_reference_str)

            def test(self):
                pass
            """)
    c.save({"conanfile.py": conanfile,
            "test_package/conanfile.py": test})
    c.run("editable add .")
    c.run("create .")
    # This used to crash due to "include" is not absolute path, now it works
    assert "Testing the package" in c.out



import os
import textwrap
from time import sleep

from conan.api.model import RecipeReference
from conan.test.utils.tools import TestClient, GenConanfile, TestServer
from conan.internal.util.files import load


def test_update_binaries():
    client = TestClient(default_server_user=True)
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.files import save, load
        import os, random
        class Pkg(ConanFile):
            def package(self):
                save(self, os.path.join(self.package_folder, "file.txt"), str(random.random()))

            def package_info(self):
                content = load(self, os.path.join(self.package_folder, "file.txt"))
                self.output.warning("CONTENT=>{}#".format(content))

        """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --name=pkg --version=0.1 --user=lasote --channel=testing")
    client.run("upload pkg/0.1@lasote/testing -r default")

    client2 = TestClient(servers=client.servers, inputs=["admin", "password"])
    client2.run("install --requires=pkg/0.1@lasote/testing")

    def get_value_from_output(output):
        tmp = str(output).split("CONTENT=>")[1]
        return tmp.split("#")[0]

    value = get_value_from_output(client2.out)

    client.run("create . --name=pkg --version=0.1 --user=lasote --channel=testing")  # Because of random, this should be NEW prev
    client.run("upload pkg/0.1@lasote/testing -r default")

    client2.run("install --requires=pkg/0.1@lasote/testing")
    new_value = get_value_from_output(client2.out)
    assert value == new_value

    client2.run("install --requires=pkg/0.1@lasote/testing --update")
    assert "Current package revision is older than the remote one" in client2.out
    new_value = get_value_from_output(client2.out)
    assert value != new_value

    # Now check newer local modifications are not overwritten
    client.run("create . --name=pkg --version=0.1 --user=lasote --channel=testing")
    client.run("upload pkg/0.1@lasote/testing -r default")

    client2.save({"conanfile.py": conanfile})
    client2.run("create . --name=pkg --version=0.1 --user=lasote --channel=testing")
    client2.run("install --requires=pkg/0.1@lasote/testing")
    value2 = get_value_from_output(client2.out)
    client2.run("install --requires=pkg/0.1@lasote/testing --update -r default")
    assert "Current package revision is newer than the remote one" in client2.out
    new_value = get_value_from_output(client2.out)
    assert value2 == new_value


def test_update_not_date():
    client = TestClient(default_server_user=True)
    # Regression for https://github.com/conan-io/conan/issues/949
    client.save({"conanfile.py": GenConanfile("hello0", "1.0")})
    client.run("export . --user=lasote --channel=stable")
    client.save({"conanfile.py": GenConanfile("hello1", "1.0").
                with_requirement("hello0/1.0@lasote/stable")},
                clean_first=True)
    client.run("install . --build='*'")
    client.run("upload hello0/1.0@lasote/stable -r default")

    prev = client.get_latest_package_reference("hello0/1.0@lasote/stable")

    ref = RecipeReference.loads("hello0/1.0@lasote/stable")

    initial_recipe_timestamp = client.cache.get_latest_recipe_revision(ref).timestamp
    initial_package_timestamp = prev.timestamp

    # Change and rebuild package
    client.save({"conanfile.py": GenConanfile("hello0", "1.0").with_class_attribute("author = 'O'")},
                clean_first=True)
    client.run("export . --user=lasote --channel=stable")
    client.run("install --requires=hello0/1.0@lasote/stable --build='*'")

    rebuild_recipe_timestamp = client.cache.get_latest_recipe_revision(ref).timestamp
    rebuild_package_timestamp = client.get_latest_package_reference(ref).timestamp

    assert rebuild_recipe_timestamp != initial_recipe_timestamp
    assert rebuild_package_timestamp != initial_package_timestamp

    # back to the consumer, try to update
    client.save({"conanfile.py": GenConanfile("hello1", "1.0").
                with_requirement("hello0/1.0@lasote/stable")}, clean_first=True)
    # First assign the preference to a remote, it has been cleared when exported locally
    client.run("install . --update")
    # *1 With revisions here is removing the package because it doesn't belong to the recipe

    client.assert_listed_require({"hello0/1.0@lasote/stable": "Newer"})

    failed_update_recipe_timestamp = client.cache.get_latest_recipe_revision(ref).timestamp
    failed_update_package_timestamp = client.get_latest_package_reference(ref).timestamp

    assert rebuild_recipe_timestamp == failed_update_recipe_timestamp
    assert rebuild_package_timestamp == failed_update_package_timestamp


def test_reuse():
    client = TestClient(default_server_user=True)
    conanfile = GenConanfile("hello0", "1.0")\
        .with_exports_sources("*")\
        .with_import("from conan.tools.files import copy")\
        .with_package("copy(self, '*', self.source_folder, self.package_folder)")
    client.save({"conanfile.py": conanfile,
                 "header.h": "content1"})
    client.run("export . --user=lasote --channel=stable")
    client.run("install --requires=hello0/1.0@lasote/stable --build='*'")
    client.run("upload hello0/1.0@lasote/stable -r default")

    client2 = TestClient(servers=client.servers, inputs=["admin", "password"])
    client2.run("install --requires=hello0/1.0@lasote/stable")
    assert "hello0/1.0@lasote/stable: Retrieving package" in client2.out

    client.save({"header.h": "//EMPTY!"})
    sleep(1)
    client.run("export . --user=lasote --channel=stable")
    client.run("install --requires=hello0/1.0@lasote/stable --build='*'")
    client.run("upload hello0/1.0@lasote/stable -r default")

    client2.run("install --requires=hello0/1.0@lasote/stable --update")
    ref = RecipeReference.loads("hello0/1.0@lasote/stable")
    pref = client.get_latest_package_reference(ref)
    package_path = client2.get_latest_pkg_layout(pref).package()
    header = load(os.path.join(package_path, "header.h"))
    assert header == "//EMPTY!"


def test_update_binaries_failed():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile()})
    client.run("create . --name=pkg --version=0.1 --user=lasote --channel=testing")
    client.run("install --requires=pkg/0.1@lasote/testing --update")
    assert "WARN: Can't update, there are no remotes defined" in client.out


def test_install_update_repeated_tool_requires():
    """
    Test that requiring the same thing multiple times, like a tool-requires, only
    require checking the servers 1, so it is much faster

    https://github.com/conan-io/conan/issues/13508
    """
    c = TestClient(default_server_user=True)
    c.save({"tool/conanfile.py": GenConanfile("tool", "0.1"),
            "liba/conanfile.py": GenConanfile("liba", "0.1"),
            "libb/conanfile.py": GenConanfile("libb", "0.1").with_requires("liba/0.1"),
            "libc/conanfile.py": GenConanfile("libc", "0.1").with_requires("libb/0.1"),
            "profile": "[tool_requires]\ntool/0.1"
            })
    c.run("create tool")
    c.run("create liba")
    c.run("create libb")
    c.run("create libc")
    c.run("install libc --update -pr=profile")
    assert 1 == str(c.out).count("tool/0.1: Checking remote")


class TestUpdateOldPolicy:
    def test_multi_remote_update_resolution(self):
        c = TestClient(servers={"r1": TestServer(), "r2": TestServer(), "r3": TestServer()},
                       inputs=["admin", "password"] * 3, light=True)
        c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
        c.run("export .")
        rev1 = c.exported_recipe_revision()
        c.run("upload * -r=r1 -c")
        # second revision
        c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_class_attribute("auther = 'me'")})
        c.run("export .")
        rev2 = c.exported_recipe_revision()
        assert rev1 != rev2
        c.run("upload * -r=r2 -c")  # By default uploads latest revisions only
        assert rev1 not in c.out
        assert rev2 in c.out
        # going back to the previous revision
        c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
        c.run("export .")  # Makes it the latest
        rev3 = c.exported_recipe_revision()
        assert rev1 == rev3
        c.run("upload * -r=r3 -c")  # By default uploads latest revisions only
        assert rev3 in c.out
        assert rev2 not in c.out

        # now test the --update, it will pick up the latest revision, which is r3
        c.run("remove * -c")
        c.run("graph info --requires=pkg/0.1 --update")
        assert f"pkg/0.1#{rev3} - Downloaded (r3)" in c.out

        # But if we enable order-based first found timestamp, it will pick up r2
        c.run("remove * -c")
        c.run("graph info --requires=pkg/0.1 --update -cc core:update_policy=legacy")
        assert "The 'core:update_policy' conf is deprecated and will be removed" in c.out
        assert f"pkg/0.1#{rev2} - Downloaded (r2)" in c.out

    def test_multi_remote_update_resolution_2_remotes(self):
        c = TestClient(servers={"r1": TestServer(), "r2": TestServer()},
                       inputs=["admin", "password"] * 2, light=True)
        c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
        c.run("export .")
        rev1 = c.exported_recipe_revision()
        c.run("upload * -r=r1 -c")
        # second revision
        c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_class_attribute("auther = 'me'")})
        c.run("export .")
        rev2 = c.exported_recipe_revision()
        assert rev1 != rev2
        c.run("upload * -r=r1 -c")
        c.run("list *#* -r=r1")

        # Upload the first, old revision to the other remote
        c.run(f"upload pkg/0.1#{rev1} -r=r2 -c")
        assert rev1 in c.out
        assert rev2 not in c.out
        c.run("list *#* -r=r2")

        # now test the --update, it will pick up the latest revision, which is r3
        c.run("remove * -c")
        c.run("graph info --requires=pkg/0.1 --update")
        assert f"pkg/0.1#{rev1} - Downloaded (r2)" in c.out

        # But if we enable order-based first found timestamp, it will pick up r2
        c.run("remove * -c")
        c.run("graph info --requires=pkg/0.1 --update -cc core:update_policy=legacy")
        assert f"pkg/0.1#{rev2} - Downloaded (r1)" in c.out

    def test_lockfile(self):
        # https://github.com/conan-io/conan/issues/18006
        c = TestClient(default_server_user=True, light=True)  # needs server to fail
        c.save_home({"global.conf": "core:update_policy=legacy"})
        c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
                "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_python_requires("dep/0.1")})
        c.run("create dep")
        c.run("lock create pkg --lockfile-out base.lock --build=* --update")
        c.run("lock create pkg --lockfile base.lock --lockfile-out full.lock --build=* --update")
        # it doesn't crash
        assert "Generated lockfile" in c.out



import time
from datetime import timedelta

import jwt
import pytest
from jwt import DecodeError

from conans.server.crypto.jwt.jwt_credentials_manager import JWTCredentialsManager


def test_jwt_manager():
    # Instance the manager to generate tokens that expires in 10 ms
    manager = JWTCredentialsManager(secret="1234asdf", expire_time=timedelta(seconds=1))

    # Encrypt a profile
    token = manager.get_token_for("myuser")

    # Decrypt the profile
    assert "myuser" == manager.get_user(token)
    with pytest.raises(DecodeError):
        manager.get_user("invalid_user")

    # Now wait 2 seconds and check if its valid now
    time.sleep(2)
    with pytest.raises(jwt.ExpiredSignatureError):
        manager.get_user(token)
